from typing import Any, NamedTuple, Union

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Integer, Shaped, jaxtyped

jax.config.update('jax_array', True)

TRUE_ARRAY: Bool[Array, ""] = lax.full((), True, dtype=jnp.bool_)
FALSE_ARRAY: Bool[Array, ""] = lax.full((), False, dtype=jnp.bool_)
INF_ARRAY: Float[Array, ""] = lax.full((), jnp.inf)

Index = Integer[Array, ""]

CanvasMask = Bool[Array, "#width #height"]
BatchCanvasMask = Bool[Array, "#batch #width #height"]
Canvas = Float[Array, "width height channel"]
ZBuffer = Float[Array, "width height"]
Colour = Float[Array, "channel"]
# triangle thus 3 "colour"s
TriangleColours = Float[Array, "3 channel"]
Vec2i = Integer[Array, "2"]
Vec3i = Integer[Array, "3"]
Vec2f = Float[Array, "2"]
Vec3f = Float[Array, "3"]
# usually used only for 3D homogeneous coordinates
Vec4f = Float[Array, "4"]
# 3 vertices, with each vertex defined in Vec2i in screen(canvas) space
Triangle2D = Integer[Array, "3 2"]
# 3 vertices, with each vertex defined in Vec2f
Triangle2Df = Float[Array, "3 2"]
# 3 vertices, each vertex defined in Vec2i in 3d (world/model) space + Float z
Triangle = Float[Array, "3 4"]
# Barycentric coordinates has 3 components
TriangleBarycentric = Float[Array, "3 3"]

# each face has 3 vertices
FaceIndices = Integer[Array, "faces 3"]
# each vertex is defined by 3 float numbers, x-y-z
Vertices = Float[Array, "vertices 3"]
Normals = Float[Array, "normals 3"]
UVCoordinates = Float[Array, "uv_counts 2"]
Texture = Float[Array, "textureWidth textureHeight channel"]
SpecularMap = Float[Array, "textureWidth textureHeight"]
NormalMap = Float[Array, "textureWidth textureHeight 3"]


class DtypeInfo(NamedTuple):
    # TODO: use Generic NamedTuple when bump to Python 3.11
    min: Union[jnp.floating[Any], jnp.integer[Any]]
    max: Union[jnp.floating[Any], jnp.integer[Any]]
    bits: int
    dtype: type

    @classmethod
    @jaxtyped
    # cannot be jitted as `dtype` will not be a valid JAX type
    def create(cls, dtype: type) -> "DtypeInfo":
        with jax.ensure_compile_time_eval():
            if jnp.issubdtype(dtype, jnp.floating):
                finfo = jnp.finfo(dtype)

                return cls(
                    min=finfo.min,
                    max=finfo.max,
                    bits=finfo.bits,
                    dtype=dtype,
                )
            if jnp.issubdtype(dtype, jnp.integer):
                iinfo = jnp.iinfo(dtype)

                return cls(
                    min=iinfo.min,
                    max=iinfo.max,
                    bits=iinfo.bits,
                    dtype=dtype,
                )

        raise ValueError(f"Unexpected dtype {dtype}")


class LightSource(NamedTuple):
    direction: Vec3f = jax.numpy.array((0., 0., -1.))
    """in world space, as it goes from that position to origin (camera).

    For example, if direction = camera's eye, full specular will be applied to
    triangles with normal towards camera.
    """
    colour: Colour = jax.numpy.ones(3)


class Buffers(NamedTuple):
    """Use lax.full to create buffers and attach here."""
    # TODO: use Generic NamedTuple when bump to Python 3.11
    zbuffer: ZBuffer
    targets: tuple[Shaped[Array, "width height ..."]]



from functools import partial
from typing import Any, NamedTuple, TypeVar, Union

import jax
import jax.lax as lax
import jax.numpy as jnp
from jax import lax
from jax.tree_util import tree_map
from jaxtyping import Array, Bool, Float, Integer, Num, jaxtyped

from .geometry import (Camera, Interpolation, barycentric, interpolate,
                       normalise_homogeneous)
from .shader import (ID, MixedExtraT, MixerOutput, PerFragment, PerVertex,
                     Shader, ShaderExtraInputT, VaryingT)
from .types import (FALSE_ARRAY, Buffers, FaceIndices, Triangle, Triangle2Df,
                    Vec2f, Vec2i, Vec3f, Vec4f)

jax.config.update('jax_array', True)


class PerVertexInScreen(NamedTuple):
    """Built-in output from Vertex Shader after Viewport transformation.

    gl_Position is in screen-space.
    """
    gl_Position: Vec4f
    # gl_PointSize is meaningful only when rendering point primitives.
    # not supported for now
    # gl_PointSize: Float[Array, ""]


class PerPrimitive(NamedTuple):
    """Input for each primitive, using outputs from Vertex Shader.

    gl_Position is in screen-space, normalised homogeneous coordinate
    """
    gl_Position: Triangle
    # gl_PointSize is meaningful only when rendering point primitives.
    # not supported for now
    # gl_PointSize: Float[Array, "primitives"]


@jaxtyped
@partial(jax.jit, static_argnames=("shader", ), donate_argnums=(1, ))
def _postprocessing(
    shader: Union[Shader[ShaderExtraInputT, VaryingT, MixedExtraT],
                  type[Shader[ShaderExtraInputT, VaryingT, MixedExtraT]]],
    buffers: Buffers,
    per_primitive: tuple[Any, ...],  # Batch PerPrimitive
    varyings: VaryingT,
    extra: ShaderExtraInputT,
) -> Buffers:
    with jax.ensure_compile_time_eval():
        # loop along first axis, for memory efficiency
        # TODO: benchmark if this is actually faster
        loop_size: int = int(buffers[0].shape[0])
        # vmap batch along second axis
        batch_size: int = int(buffers[0].shape[1])

    @jaxtyped
    @jax.jit
    def _per_pixel(coord: Vec2i) -> tuple[MixerOutput, MixedExtraT]:

        assert isinstance(coord, Vec2i), f"expected Vec2i, got {coord}"

        @jaxtyped
        def _per_primitive_process(
            primitive: PerPrimitive,
            varying_per_primitive: VaryingT,
        ) -> tuple[PerFragment, VaryingT]:
            screen: Triangle = primitive.gl_Position

            # 2d screen coordinates
            screen2d: Triangle2Df = screen[:, :2]
            assert isinstance(screen2d, Triangle2Df)

            # PROCESS: Rasterisation (Interpolate)
            # barycentric coordinates
            bc_screen: Vec3f = barycentric(screen2d, coord)
            assert isinstance(bc_screen, Vec3f)

            def _when_keep_triangle() -> tuple[PerFragment, VaryingT]:
                # weight barycentric coordinates by 1/w to obtain
                # perspective-coorected barycentric coordinates
                bc_clip: Vec3f = bc_screen * screen[:, 3]

                # normalise barycentric coordinates so that it sums to 1.
                bc_clip = bc_clip / bc_clip.sum()

                # Prepare inputs for fragment shader
                # no need to "normalise" as barycentric coordinates here
                # ensures it sums to 1, thus result is normalised when inputs
                # are.
                # Mode: NOPERSPECTIVE: since inverse of the depth is linear,
                # the correct way to interpolate it is just to interpolate
                # under screen space using bc_screen,
                # or using `NONPERSPECTVIE` mode.
                gl_FragCoord: Vec4f = interpolate(
                    screen,
                    bc_screen,
                    bc_clip,
                    mode=Interpolation.NOPERSPECTIVE,
                )
                gl_FragCoord.at[:2].set(coord)
                assert isinstance(gl_FragCoord, Vec4f)

                # Ref: https://registry.khronos.org/OpenGL-Refpages/gl4/html/gl_FrontFacing.xhtml
                gl_FrontFacing: Bool[Array, ""] = jnp.cross(
                    screen[1, :3] - screen[0, :3],
                    screen[2, :3] - screen[0, :3],
                )[-1] > 0
                assert isinstance(gl_FrontFacing, Bool[Array, ""])

                gl_PointCoord: Vec2f
                with jax.ensure_compile_time_eval():
                    # TODO: implement Point primitive properly.
                    gl_PointCoord = lax.full((2, ), 0)

                varying: VaryingT = shader.interpolate(
                    varying_per_primitive,
                    bc_screen,
                    bc_clip,
                )
                assert isinstance(varying, tuple)

                # PROCESS: Fragment Processing
                per_frag: PerFragment
                extra_fragment_output: VaryingT
                per_frag, extra_fragment_output = shader.fragment(
                    gl_FragCoord=gl_FragCoord,
                    gl_FrontFacing=gl_FrontFacing,
                    gl_PointCoord=gl_PointCoord,
                    varying=varying,
                    extra=extra,
                )
                assert isinstance(per_frag, PerFragment)
                assert isinstance(extra_fragment_output, tuple)

                # enforce default `gl_FragDepth` when it is None
                per_frag = lax.cond(
                    per_frag.use_default_depth,
                    lambda: per_frag._replace(gl_FragDepth=gl_FragCoord[2]),
                    lambda: per_frag,
                )
                assert isinstance(per_frag, PerFragment)

                return per_frag, extra_fragment_output

            # END OF `_when_keep_triangle`

            in_triangle: Bool[Array, ""] = (bc_screen >= 0).all()
            assert isinstance(in_triangle, Bool[Array, ""])

            built_in: PerFragment
            attachments: VaryingT
            built_in, attachments = lax.cond(
                in_triangle,
                _when_keep_triangle,
                # discard out-of-triangle values
                lambda: (
                    PerFragment(keeps=FALSE_ARRAY),
                    # dummy values
                    tree_map(lambda field: field[0], varying_per_primitive),
                ),
            )
            assert isinstance(built_in, PerFragment)
            assert isinstance(attachments, tuple)

            return built_in, attachments

        # END OF `_per_primitive_process`

        built_in, extra_outputs = jax.vmap(_per_primitive_process)(
            per_primitive,
            varyings,
        )
        gl_Depths = built_in.gl_FragDepth
        keeps = built_in.keeps

        # PROCESS: Per-Sample Operations (Mixing: depth test + colour blending)
        mixed_output: MixerOutput
        attachments: MixedExtraT
        mixed_output, attachments = shader.mix(gl_Depths, keeps, extra_outputs)
        assert isinstance(mixed_output, MixerOutput)
        assert isinstance(attachments, tuple)

        return mixed_output, attachments

    @jaxtyped
    @partial(jax.jit, donate_argnums=(1, ))
    def loop_body(
        index: Integer[Array, ""],
        buffers: Buffers,
    ) -> Buffers:

        _valueT = TypeVar('_valueT', bound=tuple[Any, ...])

        @jaxtyped
        @partial(jax.jit, donate_argnums=(2, ))
        def select_value_per_pixel(
            keep: Bool[Array, ""],
            new_values: _valueT,
            old_values: _valueT,
        ) -> _valueT:
            """Choose new value of the pixel, or keep the previous."""
            FieldRowT = TypeVar("FieldRowT")

            def _select_per_field(
                new_field_value: FieldRowT,
                old_field_value: FieldRowT,
            ) -> FieldRowT:
                """Choose this pixel for this field in the PyTree."""
                return lax.cond(
                    keep,
                    lambda: new_field_value,
                    lambda: old_field_value,
                )

            # tree_map over each field in the PyTree
            result: _valueT = tree_map(
                _select_per_field,
                new_values,
                old_values,
            )

            return result

        keeps: Bool[Array, "height"]
        depths: Num[Array, "height"]
        extras: MixedExtraT
        # vmap over axis 1 (height) of the buffers. Axis 0 (width) is `index`.
        (keeps, depths), extras = jax.vmap(_per_pixel)(lax.concatenate(
            (
                lax.full((batch_size, 1), index),
                lax.broadcasted_iota(int, (batch_size, 1), 0),
            ),
            1,
        ))
        assert isinstance(keeps, Bool[Array, "height"])
        assert isinstance(depths, Num[Array, "height"])
        assert isinstance(extras, tuple)

        # vmap each pixel over axis 1 (height) of the buffers (per row in
        # matrix)
        buffers_row = jax.vmap(select_value_per_pixel)(
            keeps,
            Buffers(zbuffer=depths, targets=tuple(extras)),
            tree_map(lambda field: field[index], buffers),
        )

        # tree_map over each field in the PyTree to update all buffers
        return tree_map(
            lambda field, value: field.at[index].set(value),
            buffers,
            buffers_row,
        )

    # END OF `loop_body`

    # iterate over axis 0 (width) of the buffers (one row at a time)
    buffers = lax.fori_loop(
        0,
        loop_size,
        loop_body,
        buffers,
    )
    assert isinstance(buffers, Buffers)

    return buffers


@jaxtyped
@partial(jax.jit, static_argnames=("shader", ), donate_argnums=(2, ))
def render(
    camera: Camera,
    shader: Union[Shader[ShaderExtraInputT, VaryingT, MixedExtraT],
                  type[Shader[ShaderExtraInputT, VaryingT, MixedExtraT]]],
    buffers: Buffers,
    face_indices: FaceIndices,
    extra: ShaderExtraInputT,
) -> Buffers:
    vertices_count: int
    gl_InstanceID: ID
    with jax.ensure_compile_time_eval():
        vertices_count = extra[0].shape[0]
        gl_InstanceID = jnp.array(0, dtype=int)
        assert isinstance(vertices_count, int)
        assert isinstance(gl_InstanceID, ID)

    @jaxtyped
    @jax.jit
    def vertex_processing(
            gl_VertexID: Integer[Array, ""],  #
    ) -> tuple[PerVertexInScreen, VaryingT]:
        """Process one vertex into screen space, and keep varying values."""
        per_vertex: PerVertex
        varying: VaryingT
        per_vertex, varying = shader.vertex(
            gl_VertexID,
            gl_InstanceID,
            camera,
            extra,
        )
        assert isinstance(per_vertex, PerVertex)
        assert isinstance(varying, tuple)

        # TODO: add clipping in clip space.

        # NDC, normalised device coordinate
        # Ref: OpenGL Spec 4.6 (Core Profile), Section 15.2.2
        w: Float[Array, ""] = per_vertex.gl_Position[3]
        ndc: Vec4f = normalise_homogeneous(per_vertex.gl_Position)
        assert isinstance(ndc, Vec4f)

        # already normalised; result is still normalised
        # Ref: OpenGL Spec 4.6 (Core Profile), Section 15.2.2
        screen: Vec4f = (camera.viewport @ ndc).at[3].divide(w)
        assert isinstance(screen, Vec4f)

        return PerVertexInScreen(gl_Position=screen), varying

    # PROCESS: Vertex Processing
    per_vertices, varyings = jax.vmap(vertex_processing)(
        lax.iota(int, vertices_count),  # gl_VertexID
    )

    # everything after vertex processing, will directly update buffers
    buffers = _postprocessing(
        shader=shader,
        buffers=buffers,
        per_primitive=tree_map(lambda field: field[face_indices],
                               per_vertices),
        varyings=tree_map(lambda field: field[face_indices], varyings),
        extra=extra,
    )
    assert isinstance(buffers, Buffers)

    return buffers



import enum
from functools import partial
from typing import Any, NamedTuple, Optional

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Float, Integer, Num, jaxtyped

from .types import Triangle2Df, Vec2f, Vec3f, Vec4f

# Transform matrix that takes a batch of homogeneous 3D vertices and transform
# them into 2D cartesian vertices in screen space + Z value (making it 3D)
#
# The result of x-y values in screen space may be float, and thus further
# conversion to integers are needed.
World2Screen = Float[Array, "4 4"]
# Transform all coordinates from world space to view space, with camera at
# origin. (Object Coordinates -> Eye Coordinates)
View = Float[Array, "4 4"]
# Transform all coordinates from view space to viewing volume.
# (Eye Coordinates -> Clip Coordinates)
Projection = Float[Array, "4 4"]
# Transform all coordinates from clip space in a bi-unit cube ([-1...1]^3) to
# a screen cube ([x, x+width] * [y, y+height] * [0, depth]) in view space.
# (Normalised Device Coordinates -> Window Coordinates)
Viewport = Float[Array, "4 4"]


@jaxtyped
@partial(jax.jit, donate_argnums=(0, ))
def normalise(vector: Float[Array, "dim"]) -> Float[Array, "dim"]:
    """normalise vector in-place."""
    return vector / jnp.linalg.norm(vector)


class Interpolation(enum.Enum):
    """Interpolation methods for rasterisation.

    References:
      - [Interpolation qualifiers](https://www.khronos.org/opengl/wiki/Type_Qualifier_(GLSL)#Interpolation_qualifiers)
    """
    FLAT = 0
    """Flat shading: use the value of the first vertex of the primitive""" ""
    NOPERSPECTIVE = 1
    """No perspective correction: linear interpolation in screen space"""
    SMOOTH = 2
    """Perspective correction: linear interpolation in clip space"""

    @jaxtyped
    @partial(jax.jit, static_argnames=("self", ))
    def __call__(
        self,
        values: Num[Array, "3 *valueDimensions"],
        barycentric_screen: Vec3f,
        barycentric_clip: Vec3f,
    ) -> Num[Array, "*valueDimensions"]:
        """Interpolation, using barycentric coordinates.

        Parameters:
          - values: values at the vertices of the triangle, with axis 0 being
            the batch axis.
          - barycentric_screen: barycentric coordinates in screen space of the
            point to interpolate
          - barycentric_clip: barycentric coordinates in clip space of the
            point to interpolate
        """
        dtype = jax.dtypes.result_type(
            barycentric_screen,
            barycentric_clip,
            values,
        )
        coef: Vec3f
        # branches are ok because `self` is static: decided at compile time
        if self == Interpolation.FLAT:
            with jax.ensure_compile_time_eval():
                coef = jnp.array([1, 0, 0], dtype=dtype)
        elif self == Interpolation.NOPERSPECTIVE:
            coef = barycentric_screen
        elif self == Interpolation.SMOOTH:
            coef = barycentric_clip
        else:
            raise ValueError(f"Unknown interpolation method {self}")

        interpolated = lax.dot_general(
            coef.astype(dtype),
            values.astype(dtype),
            (((0, ), (0, )), ([], [])),
        )

        return interpolated


@jaxtyped
@partial(jax.jit, static_argnames=("mode", ))
def interpolate(
    values: Num[Array, "3 *valueDimensions"],
    barycentric_screen: Vec3f,
    barycentric_clip: Vec3f,
    mode: Interpolation = Interpolation.SMOOTH,
) -> Num[Array, "*valueDimensions"]:
    """Convenient wrapper, see `Interpolation.__call__`.

    Default mode is `Interpolation.SMOOTH`.
    """
    interpolated: Num[Array, "*valueDimensions"]
    interpolated = mode(
        values=values,
        barycentric_screen=barycentric_screen,
        barycentric_clip=barycentric_clip,
    )
    assert isinstance(interpolated, Num[Array, "*valueDimensions"])

    return interpolated


@jaxtyped
@jax.jit
def to_homogeneous(
    coordinates: Float[Array, "*batch dim"],
    value: Float[Array, "*batch"] = jnp.array(1.),
) -> Float[Array, "*batch dim+1"]:
    """Transform the coordinates to homogeneous coordinates by append a batch
        of `value`s (default 1.) in the last axis."""
    if not isinstance(value, Float[Array, "*batch"]):
        value = jnp.array(value)

    paddings: Float[Array, "*batch 1"] = jnp.broadcast_to(
        value.astype(jax.dtypes.result_type(coordinates)),
        (*coordinates.shape[:-1], 1),
    )
    homo_coords: Float[Array, "*batch dim+1"] = lax.concatenate(
        (coordinates, paddings),
        jnp.ndim(coordinates) - 1,
    )

    return homo_coords


@jaxtyped
@jax.jit
def normalise_homogeneous(
    coordinates: Float[Array, "*batch dim"], ) -> Float[Array, "*batch dim"]:
    """Transform the homogenous coordinates to make the scale factor equals to
        either 1 or 0, by divide every element with the last element on the
        last axis.

    Noted that when a coordinate is 0 and divides by 0, it will produce a nan;
    for non-zero elements divides by 0, a inf will be produced.
    """
    return coordinates / coordinates[..., -1:]


@jaxtyped
@jax.jit
def to_cartesian(
    coordinates: Float[Array, "*batch dim"], ) -> Float[Array, "*batch dim-1"]:
    """Transform the homogenous coordinates to cartesian coordinates by divide
        every element with the last element on the last axis, then drop them.

    When last component is 0, this function just discard the w-component
    without division.
    """
    return jnp.where(
        # if w component is 0, just discard it and return.
        coordinates[..., -1:] == .0,
        coordinates[..., :-1],
        normalise_homogeneous(coordinates)[..., :-1],
    )


class Camera(NamedTuple):
    """Camera parameters.

    - view: transform from model space to view space
    - projection: transform from view space to clip space
    - viewport: transform from NDC (normalised device coordinate) space to
      screen space. Noticed that this is NDC space in OpenGL, which has range
      [-1, 1]^3.
    - world_to_clip: transform from model space to clip space
    - world_to_eye_norm: transform normals from model space to eye space, without projection.
    - world_to_screen: transform from model space to screen space
    """
    view: View
    projection: Projection
    viewport: Viewport
    world_to_clip: Projection
    world_to_eye_norm: Projection
    world_to_screen: World2Screen
    view_inv: View
    screen_to_world: World2Screen

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ))
    def create(
        cls,
        view: View,
        projection: Projection,
        viewport: Viewport,
        view_inv: Optional[View] = None,
    ) -> "Camera":
        """Create a camera with the given parameters.

        Parameters:
          - view: transform from model space to view space
          - projection: transform from view space to clip space
          - viewport: transform from NDC (normalised device coordinate) space to
          - view_inv: inverse of view. If not given, it will be computed.
        """
        if view_inv is None:
            view_inv = jnp.linalg.inv(view)
        assert isinstance(view_inv, View)

        projection_inv: Projection = lax.cond(
            jnp.isclose(projection[3, 3], 0),
            # is perspective projection matrix
            cls.perspective_projection_matrix_inv,
            # is orthographic projection matrix
            cls.orthographic_projection_matrix_inv,
            # arg
            projection,
        )
        assert isinstance(projection_inv, Projection), f"{projection_inv}"

        viewport_inv: Viewport = cls.viewport_matrix_inv(viewport)
        assert isinstance(viewport_inv, Viewport)

        return cls(
            view=view,
            viewport=viewport,
            projection=projection,
            world_to_clip=projection @ view,
            # inverse transpose of projection @ view
            world_to_eye_norm=view_inv.T,
            world_to_screen=viewport @ projection @ view,
            view_inv=view_inv,
            screen_to_world=view_inv @ projection_inv @ viewport_inv,
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def apply(
        points: Num[Array, "*N 4"],
        matrix: Num[Array, "4 4"],
    ) -> Num[Array, "*N 4"]:
        """Transform homogeneous points using given matrix.

        Parameters:
          - points: shape (4, ) or (N, 4). points in model space, with axis 0
            being the batch axis. Batch axis can be omitted. Points must be
            in homogeneous coordinate.
          - matrix: shape (4, 4) transformation matrix

        Returns: coordinates transformed
        """
        assert jnp.ndim(points) < 3
        assert (((jnp.ndim(points) == 2) and (points.shape[1] == 4))
                or ((jnp.ndim(points) == 1) and (points.shape[0] == 4)))

        with jax.ensure_compile_time_eval():
            lhs_contract_axis = 1 if jnp.ndim(points) == 2 else 0
            dtype = jax.dtypes.result_type(points, matrix)

        # put `points` at lhs to keep batch axis at axis 0 in the result.
        transformed: Num[Array, "*N 4"] = lax.dot_general(
            points.astype(dtype),
            matrix.astype(dtype),
            (((lhs_contract_axis, ), (1, )), ([], [])),
        )
        assert isinstance(transformed, Num[Array, "*N 4"])

        return transformed

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ))
    def apply_pos(
        cls,
        points: Num[Array, "*N 3"],
        matrix: Num[Array, "4 4"],
    ) -> Num[Array, "*N 3"]:
        """Transform points representing 3D positions using given matrix.

        Parameters:
          - points: shape (3, ) or (N, 3). points in model space, with axis 0
            being the batch axis. Batch axis can be omitted. Points must be
            in cartesian coordinate. For coordinates in homogeneous coordinate,
            use `apply` instead.
          - matrix: shape (4, 4) transformation matrix

        Returns: coordinates transformed
        """
        points_homo = to_homogeneous(points)
        assert isinstance(points_homo, Num[Array, "*N 4"])

        transformed_homo = cls.apply(points_homo, matrix)
        assert isinstance(transformed_homo, Num[Array, "*N 4"])

        transformed = to_cartesian(transformed_homo)
        assert isinstance(transformed, Num[Array, "*N 3"])

        return transformed

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ))
    def apply_vec(
        cls,
        vectors: Num[Array, "*N 3"],
        matrix: Num[Array, "4 4"],
    ) -> Num[Array, "*N 3"]:
        """Transform vectors representing 3D positions using given matrix.

        Parameters:
          - vectors: shape (3, ) or (N, 3). Directional Vectors in model
            space, with axis 0 being the batch axis. Batch axis can be omitted.
            Vectors must be in cartesian coordinate. For coordinates in
            homogeneous coordinate, use `apply` instead.
          - matrix: shape (4, 4) transformation matrix

        Returns: vectors transformed and normalised
        """
        normalised_vectors = normalise(vectors)
        assert isinstance(normalised_vectors, Num[Array, "*N 3"])

        points_homo = to_homogeneous(
            normalised_vectors,
            jnp.zeros((), dtype=vectors.dtype),
        )
        assert isinstance(points_homo, Num[Array, "*N 4"])

        transformed_homo = cls.apply(points_homo, matrix)
        assert isinstance(transformed_homo, Num[Array, "*N 4"])

        transformed = transformed_homo[..., :3]
        assert isinstance(transformed, Num[Array, "*N 3"])

        transformed_normalised = normalise(transformed)
        assert isinstance(transformed_normalised, Num[Array, "*N 3"])

        return transformed_normalised

    @jaxtyped
    @jax.jit
    def to_screen(
        self,
        points: Num[Array, "*N 4"],
    ) -> Num[Array, "*N 4"]:
        """Transform points from model space to screen space.

        Parameters:
          - points: shape (4, ) or (N, 4). points in model space, with axis 0
            being the batch axis. Batch axis can be omitted. Points must be
            in homogeneous coordinate.

        Returns: points in screen space, with axis 0 being the batch axis, if
            given in batch. The dtype may be promoted. The homogeneous
            coordinates are normalised.
        """
        screen_space = self.apply(points, self.world_to_screen)
        assert isinstance(screen_space, Num[Array, "*N 4"])

        normalised = normalise_homogeneous(screen_space)
        assert isinstance(normalised, Num[Array, "*N 4"])

        return normalised

    @jaxtyped
    @jax.jit
    def to_clip(
        self,
        points: Num[Array, "*N 4"],
    ) -> Num[Array, "*N 4"]:
        """Transform points from model space to screen space.

        Parameters:
          - points: shape (4, ) or (N, 4). points in model space, with axis 0
            being the batch axis. Batch axis can be omitted. Points must be
            in homogeneous coordinate.

        Returns: points in clip space, with axis 0 being the batch axis, if
            given in batch. The dtype may be promoted. The homogeneous
            coordinates are not normalized.
        """
        clip_space = self.apply(points, self.world_to_clip)
        assert isinstance(clip_space, Num[Array, "*N 4"])

        return clip_space

    @jaxtyped
    @jax.jit
    def to_screen_inv(
        self,
        screen: Float[Array, "*N 4"],
    ) -> Float[Array, "*N 4"]:
        """Transform points from screen space to model space.

        This is an inverse process of `to_screen`, and provide higher precision
        then just multiplying the inverse. This may help solve NaN issue.

        Internally this is done by two `lax.linalg.triangular_solve` for
        viewport and projection, then a `@` for `view_inv`. If a good
        `view_inv` is provided during creation of this camera using
        `view_matrix_inv`, this should provide a much higher precision.
        """
        if screen.ndim == 1:
            _screen = screen[None, :]
        else:
            _screen = screen

        clip = lax.linalg.triangular_solve(self.viewport, _screen)
        assert isinstance(clip, Float[Array, "*N 4"])
        shuffle = lax.cond(
            self.projection[3, 3] == 0,
            # perspective projection
            lambda: jnp.array([0, 1, 3, 2]),
            # orthographic projection
            lambda: jnp.array([0, 1, 2, 3]),
        )
        eye = lax.linalg.triangular_solve(
            self.projection[..., shuffle],
            clip[..., shuffle],
        )[..., shuffle]
        assert isinstance(eye, Float[Array, "*N 4"])
        world = self.apply(eye, self.view_inv)
        assert isinstance(world, Float[Array, "*N 4"])

        if screen.ndim == 1:
            world = world[0]

        jax.debug.print(
            "s {} c {} e {} w {}",
            normalise_homogeneous(screen),
            normalise_homogeneous(clip),
            normalise_homogeneous(eye),
            normalise_homogeneous(world),
        )

        return world

    @staticmethod
    @jaxtyped
    @jax.jit
    def inv_scale_translation_matrix(
            scale_translation_mat: Float[Array, "4 4"]) -> Float[Array, "4 4"]:
        """Compute the inverse matrix of a (4, 4) matrix representing a scale and translation, in a form of:

            [[s_x, 0,   0,   t_x],
             [0,   s_y, 0,   t_y],
             [0,   0,   s_z, t_z],
             [0,   0,   0,   1]]

        where s is a scale vector and t is a translation vector. It is treated
        as a combination of a scale matrix and a translation matrix, as
        `scale @ translation`: translate first, then scale.

        This utilise the fact that the inverse of a scale operation is just the
        reciprocal of the scale factor, and the inverse of a translation is
        just the negative of the translation. It separates the scale and
        translation operations first, inverse them separately, then combine
        them back (in reverse order).
        """

        scale_inv = jnp.diag(1. / jnp.diag(scale_translation_mat))
        assert isinstance(scale_inv, Float[Array, "4 4"])

        # scale_translation = scale @ translation;
        # thus  translation = scale_inv @ scale @ translation
        #                   = scale_inv @ scale_translation
        translation: Float[Array, "4 4"] = scale_inv @ scale_translation_mat
        assert isinstance(translation, Float[Array, "4 4"])

        # inverse of translation: negative of translation
        translation_inv = (jnp.identity(4).at[:3, 3].set(-translation[:3, 3]))
        assert isinstance(translation_inv, Float[Array, "4 4"])

        scale_translation_inv = translation_inv @ scale_inv
        assert isinstance(scale_translation_inv, Float[Array, "4 4"])

        return scale_translation_inv

    @staticmethod
    @jaxtyped
    @jax.jit
    def view_matrix(
        eye: Vec3f,
        centre: Vec3f,
        up: Vec3f,
    ) -> View:
        """Compute View matrix as defined by OpenGL / tinyrenderer.

        Same as `lookAt` in OpenGL / tinyrenderer.

        Parameters:
          - eye: the position of camera, in world space
          - centre: the centre of the frame, where the camera points to, in
            world space
          - up: the direction vector with start point at "eye", indicating the
            "up" direction of the camera frame.

        Reference:
          - [gluLookAt](https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/gluLookAt.xml)
          - [glTranslate](https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/glTranslate.xml)
          - [GluLookAt Code](https://www.khronos.org/opengl/wiki/GluLookAt_code)
        """
        forward: Vec3f = normalise(centre - eye)
        up = normalise(up)
        side: Vec3f = normalise(jnp.cross(forward, up))
        up = jnp.cross(side, forward)

        m: View = (
            jnp.identity(4)  #
            .at[0, :3].set(side)  #
            .at[1, :3].set(up)  #
            .at[2, :3].set(-forward)  #
        )
        translation: View = jnp.identity(4).at[:3, 3].set(-eye)

        view: View = m @ translation

        return view

    @staticmethod
    @jaxtyped
    @jax.jit
    def view_matrix_inv(
        eye: Vec3f,
        centre: Vec3f,
        up: Vec3f,
    ) -> View:
        """Compute the invert of View matrix as defined by OpenGL.

        Same as inverting `lookAt` in OpenGL, but more precise.

        Parameters:
          - eye: the position of camera, in world space
          - centre: the centre of the frame, where the camera points to, in
            world space
          - up: the direction vector with start point at "eye", indicating the
            "up" direction of the camera frame.

        Noticed that the view matrix contains only rotation and translation, and
        thus the inverse of it is just the inverse of translation multiplies the
        inverse (a simple transpose!) of rotation.

        Returns: View^{-1}, (4, 4) matrix.

        Reference:
          - [gluLookAt](https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/gluLookAt.xml)
          - [4B](http://graphics.stanford.edu/courses/cs248-98-fall/Final/q4.html)
        """
        forward: Vec3f = normalise(centre - eye)
        up = normalise(up)
        side: Vec3f = normalise(jnp.cross(forward, up))
        up = jnp.cross(side, forward)

        # inverse of rotation is just the transpose
        m: View = (
            jnp.identity(4)  #
            .at[0, :3].set(side)  #
            .at[1, :3].set(up)  #
            .at[2, :3].set(-forward)  #
        )
        m_inv: View = m.T
        assert isinstance(m_inv, View)

        # inverse of translation is just the negative of translation
        translation_inv: View = jnp.identity(4).at[:3, 3].set(eye)
        assert isinstance(translation_inv, View)

        view_matrix_inv: View = translation_inv @ m_inv
        assert isinstance(view_matrix_inv, View)

        return view_matrix_inv

    @staticmethod
    @jaxtyped
    @jax.jit
    def perspective_projection_matrix(
        fovy: jnp.floating[Any],
        aspect: jnp.floating[Any],
        z_near: jnp.floating[Any],
        z_far: jnp.floating[Any],
    ) -> Projection:
        """Create a projection matrix to map the model in the camera frame (eye
            coordinates) onto the viewing volume (clip coordinates), using
            perspective transformation. This follows the implementation in
            OpenGL (gluPerspective)

        Parameters:
          - fovy: Specifies the field of view angle, in degrees, in the y
            direction.
          - aspect: Specifies the aspect ratio that determines the field of
            view in the x direction. The aspect ratio is the ratio of x (width)
            to y (height).
          - z_near: Specifies the distance from the viewer to the near clipping
            plane (always positive).
          - z_far: Specifies the distance from the viewer to the far clipping
            plane (always positive).

        Return: Projection, (4, 4) matrix.

        Reference:
          - [gluPerspective](https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/gluPerspective.xml)
        """
        f: jnp.single = 1. / lax.tan(
            jnp.radians(jnp.asarray(fovy).astype(jnp.single)) / 2.)
        projection: Projection = (
            jnp.zeros((4, 4), dtype=jnp.single)  #
            .at[0, 0].set(f / aspect)  #
            .at[1, 1].set(f)  #
            .at[2, 2].set((z_far + z_near) / (z_near - z_far))  #
            # translate z
            .at[2, 3].set((2. * z_far * z_near) / (z_near - z_far))  #
            .at[3, 2].set(-1.)  # let \omega be -z
        )

        return projection

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ))
    def perspective_projection_matrix_inv(cls, mat: Projection) -> Projection:
        """Create the inverse of a perspective projection matrix as defined in
            `perspective_projection_matrix`.

        Since the perspective projection matrix is formed as:

            [[a, 0,  0, 0],
             [0, b,  0, 0],
             [0, 0,  c, d],
             [0, 0, -1, 0]]

        it can be simply transformed into a scale-translation matrix by
        swapping the last two columns. Thus, the inverse is computed by
        swapping the last columns, then inverting using
        `inv_scale_translation_matrix`, and finally swapping back (last two
        rows, instead).
        """
        with jax.ensure_compile_time_eval():
            shuffle: Integer[Array, "4"] = jnp.array((0, 1, 3, 2))
            assert isinstance(shuffle, Integer[Array, "4"])

        inv = cls.inv_scale_translation_matrix(mat[:, shuffle])[shuffle, :]
        assert isinstance(inv, Projection)

        return inv

    @staticmethod
    @jaxtyped
    @jax.jit
    def orthographic_projection_matrix(
        left: jnp.floating[Any],
        right: jnp.floating[Any],
        bottom: jnp.floating[Any],
        top: jnp.floating[Any],
        z_near: jnp.floating[Any],
        z_far: jnp.floating[Any],
    ) -> Projection:
        """Create a projection matrix to map the model in the camera frame (eye
            coordinates) onto the viewing volume (clip coordinates), using
            orthographic transformation. This follows the implementation in
            OpenGL (glOrtho).

        Parameters:
          - left, right: Specifies the coordinates for the left and right
            vertical clipping planes.
          - bottom, top: Specifies the coordinates for the bottom and top
            horizontal clipping planes..
          - z_near, z_far: Specifies the distances from the viewer to the
            nearer and farther depth clipping planes. These values are negative
            if they are behind the viewer.

        Return: Projection, (4, 4) matrix.

        Reference:
          - [glOrtho](https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/glOrtho.xml)
        """
        l_op: Float[Array, "3"] = jnp.array([right, top, z_far])
        r_op: Float[Array, "3"] = jnp.array([left, bottom, z_near])
        projection: Projection = (
            jnp.zeros((4, 4), dtype=jnp.single)  #
            .at[0, 0].set(2 / (right - left))  #
            .at[1, 1].set(2 / (top - bottom))  #
            .at[2, 2].set(-2 / (z_far - z_near))  #
            .at[3, 3].set(1)  #
            .at[:3, 3].set(-(l_op + r_op) / (l_op - r_op))  #
        )

        return projection

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ))
    def orthographic_projection_matrix_inv(cls, mat: Projection) -> Projection:
        """Create the inverse of a orthographic projection matrix as defined in
            `orthographic_projection_matrix`. Since orthographic projection
            matrix is a scale-translation matrix, the inverse is computed by
            `inv_scale_translation_matrix` directly.
        """
        inv = cls.inv_scale_translation_matrix(mat)
        assert isinstance(inv, Projection)

        return inv

    @staticmethod
    @jaxtyped
    @jax.jit
    def perspective_projection_matrix_tinyrenderer(
        eye: Vec3f,
        centre: Vec3f,
        dtype: jnp.dtype = jnp.single,
    ) -> Projection:
        """Create a projection matrix to map the model in the camera frame (eye
            coordinates) onto the viewing volume (clip coordinates), using
            perspective transformation.

        Parameters:
          - eye: the position of camera, in world space
          - centre: the centre of the frame, where the camera points to, in
            world space
          - dtype: the dtype for the projection matrix.

        Return: Projection, (4, 4) matrix.
        """
        projection: Projection = (
            jnp.identity(4, dtype=dtype)  #
            .at[3, 2].set(-1 / jnp.linalg.norm(eye - centre))  #
        )

        return projection

    @staticmethod
    @jaxtyped
    @jax.jit
    def viewport_matrix(
        lowerbound: Num[Array, "2"],
        dimension: Integer[Array, "2"],
        depth: Num[Array, ""],
    ) -> Viewport:
        """Create a viewport matrix to map the model in bi-unit cube
            ([-1...1]^3) onto the screen cube ([x, x+w]*[y, y+h]*[0, d]). The
            result matrix is the viewport matrix as defined in OpenGL /
            tinyrenderer.

        Parameters:
          - lowerbound: x-y of the lower left corner of the viewport, in screen
            space.
          - dimension: width, height of the viewport, in screen space.
          - depth: the depth of the viewport in screen space, for zbuffer
          - dtype: the dtype for the viewport matrix.

        Return: Viewport, (4, 4) matrix.
        """
        width, height = dimension
        viewport: Viewport = (
            jnp.identity(4)  #
            .at[:2, 3].set(lowerbound + dimension / 2)  #
            .at[0, 0].set(width / 2).at[1, 1].set(height / 2)  #
            .at[2, 2:].set(depth / 2)  #
        )

        return viewport

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ))
    def viewport_matrix_inv(cls, viewport: Viewport) -> Viewport:
        """Create the inverse of a viewport matrix as defined in `viewport_matrix`.

        Parameters:
          - viewport: Viewport matrix to invert.

        Return: Viewport^{-1}, (4, 4) matrix.
        """
        viewport_inv: Viewport = cls.inv_scale_translation_matrix(viewport)
        assert isinstance(viewport_inv, Viewport)

        return viewport_inv

    @staticmethod
    @jaxtyped
    @jax.jit
    def world_to_screen_matrix(width: int, height: int) -> World2Screen:
        """Generate the projection matrix to convert model coordinates to
            screen/canvas coordinates.

        It assumes all model coordinates are in [-1...1] and will transform them
        into ([0...width], [0...height], [-1...1]).

        Return: World2Screen (Float[Array, "4 4"])
        """
        world2screen: World2Screen = (
            # 3. div by half to centering
            jnp.identity(4).at[0, 0].set(.5).at[1, 1].set(.5)
            # 2. mul by width, height
            @ jnp.identity(4).at[0, 0].set(width).at[1, 1].set(height)
            # 1. Add by 1 to make values positive
            @ jnp.identity(4).at[:2, -1].set(1))

        return world2screen


@jaxtyped
@jax.jit
def compute_normal(triangle_verts: Float[Array, "3 3"]) -> Float[Array, "3"]:
    normal: Float[Array, "3"] = jnp.cross(
        triangle_verts[2, :] - triangle_verts[0, :],
        triangle_verts[1, :] - triangle_verts[0, :],
    )
    normal = normal / jnp.linalg.norm(normal, keepdims=True)
    assert isinstance(normal, Float[Array, "3"])

    return normal


@jaxtyped
@jax.jit
def compute_normals(batch_verts: Float[Array, "b 3 3"]) -> Float[Array, "b 3"]:
    return jax.vmap(compute_normal)(batch_verts)


@jaxtyped
@jax.jit
def transform_matrix_from_rotation(rotation: Vec4f) -> Float[Array, "3 3"]:
    """Generate a transform matrix from a quaternion rotation.

    Supports non-unit rotation.

    References:
          - [Quaternions and spatial rotation#Quaternion-derived rotation matrix](https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Quaternion-derived_rotation_matrix)
          - [TinySceneRenderer::set_object_orientation](https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/tinyrenderer.cpp#LL997C20-L997C20)
    """
    d = rotation @ rotation
    s = 2.0 / d  # here s is $2\times s$ in Wikipedia.

    rs: Vec3f = rotation[:3] * s
    ((wx, wy, wz), (xx, xy, xz), (yy, yz, zz)) = jnp.outer(
        rotation[jnp.array((3, 0, 1))],
        rs,
    )

    mat: Float[Array, "3 3"] = jnp.array((
        (1. - (yy + zz), xy - wz, xz + wy),
        (xy + wz, 1. - (xx + zz), yz - wx),
        (xz - wy, yz + wx, 1. - (xx + yy)),
    ))
    assert isinstance(mat, Float[Array, "3 3"])

    return mat


@jaxtyped
@jax.jit
def barycentric(pts: Triangle2Df, p: Vec2f) -> Vec3f:
    """Compute the barycentric coordinate of `p`.
        Returns u[-1] < 0 if `p` is outside of the triangle.
    """
    mat: Float[Array, "3 2"] = jnp.vstack((
        pts[2] - pts[0],
        pts[1] - pts[0],
        pts[0] - p,
    ))
    v: Vec3f = jnp.cross(mat[:, 0], mat[:, 1])
    # `u[2]` is 0, that means triangle is degenerate, in this case
    # return something with negative coordinates
    v = lax.cond(
        jnp.abs(v[-1]) < 1e-10,
        lambda: jnp.array((-1., 1., 1.)),
        lambda: jnp.array((
            1 - (v[0] + v[1]) / v[2],
            v[1] / v[2],
            v[0] / v[2],
        )),
    )
    assert isinstance(v, Vec3f)

    return v



from .geometry import Camera, normalise
from .model import Model, ModelObject
from .renderer import (CameraParameters, LightParameters, Renderer,
                       ShadowParameters)
from .scene import Scene, UpAxis
from .types import Buffers, Colour, SpecularMap, Texture, Vec3f
from .utils import build_texture_from_PyTinyrenderer, transpose_for_display



from functools import partial
from typing import Any, NamedTuple, Optional, Union

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Integer, Num, jaxtyped

from .geometry import Camera, View, Projection, Viewport, normalise
from .model import MergedModel, ModelMatrix, ModelObject
from .pipeline import render
from .shaders.phong_reflection import (PhongReflectionTextureExtraInput,
                                       PhongReflectionTextureShader)
from .shaders.phong_reflection_shadow import (
    PhongReflectionShadowTextureExtraInput, PhongReflectionShadowTextureShader)
from .shadow import Shadow
from .types import (Buffers, Canvas, Colour, DtypeInfo, LightSource, Vec3f,
                    Vertices, ZBuffer)

DoubleSidedFaces = Bool[Array, "faces"]
"""Whether to render both sides of each face (triangle primitive)."""
ObjectsT = Union[list[ModelObject], tuple[ModelObject, ...]]
"""A list or tuple of model objects. TODO: Change to `PyTree[ModelObject]`."""


class CameraParameters(NamedTuple):
    """Parameters for rendering from camera.

    Default values come from [erwincoumans/tinyrenderer::TinyRendererCamera](https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/python/pytinyrenderer.cc#L56)
    """
    viewWidth: int = 640
    """width of the viewport."""
    viewHeight: int = 480
    """height of the viewport."""
    viewDepth: float = 1.
    """depth of the rendered view."""
    near: float = 0.01
    """near clipping plane."""
    far: float = 1000.
    """far clipping plane."""
    hfov: float = 58.
    """horizontal field of view, in degrees."""
    vfov: float = 45.
    """vertical field of view, in degrees."""
    position: Vec3f = jnp.ones(3)
    """position of the camera in world space."""
    target: Vec3f = jnp.zeros(3)
    """target of the camera."""
    up: Vec3f = jnp.array((0., 0., 1.))
    """up direction of the camera."""


class LightParameters(NamedTuple):
    """Parameters for directional light in rendering.

    Default values come from [erwincoumans/tinyrenderer::TinyRenderLight](https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/python/pytinyrenderer.cc#L74).
    """
    direction: Vec3f = normalise(jnp.array((0.57735, 0.57735, 0.57735)))
    """in world space, as it goes from that position to origin (camera).

    For example, if direction = camera's eye, full specular will be applied to
    triangles with normal towards camera.
    """
    colour: Colour = jnp.ones(3)
    """Light source to render."""
    ambient: Colour = jnp.array((0.6, 0.6, 0.6))
    """Ambient colour. This is added to the final colour of the object."""
    diffuse: Colour = jnp.array((0.35, 0.35, 0.35))
    """Diffuse coefficient per colour channel. This is multiplied to the
        diffuse texture colour of the object.
    """
    specular: Colour = jnp.array((0.05, 0.05, 0.05))
    """Specular coefficient per colour channel. This is multiplied to the
        computed specular light colour.
    """


class ShadowParameters(NamedTuple):
    """Parameters for rendering shadow map.

    Default values come from:
      - [erwincoumans/tinyrenderer::TinyRenderLight](https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/python/pytinyrenderer.cc#L74).
      - for `up`, [erwincoumans/tinyrenderer::renderObject](https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/tinyrenderer.cpp#L372).
    """
    centre: Vec3f = jnp.zeros(3)
    """centre of the scene, same as object's camera's centre."""
    up: Vec3f = jnp.array((0., 0., 1.))
    """up direction of the scene, same as object's camera's up."""
    strength: Colour = 1 - jnp.array((0.4, 0.4, 0.4))
    """Strength of shadow. Must be in [0, 1]. 0 means no shadow, 1 means fully
        black shadow.  See `Shadow.strength` for more details.
    """
    offset: float = 0.05
    """Offset to avoid self-shadowing / z-fighting. This will be subtracted to
        the shadow map, making the shadows further away from the light.
    """


class Renderer:

    @staticmethod
    @jaxtyped
    @partial(jax.jit, inline=True)
    def merge_objects(objects: ObjectsT) -> MergedModel:
        """Merge objects into a single model.

        Parameters:
          - objects: a list of objects to merge.

        Returns: A model containing the merged objects into one single mesh.
        """
        with jax.ensure_compile_time_eval():
            models = [obj.model for obj in objects]

            # broadcasted per vertex info
            counts: list[int] = [len(m.verts) for m in models]

            map_indices: Integer[Array, "vertices"]
            map_indices = MergedModel.generate_object_vert_info(
                counts,
                list(range(len(models))),
            )
            assert isinstance(map_indices, Integer[Array, "vertices"])

            map_wh_per_object = jnp.asarray(
                [m.diffuse_map.shape[:2] for m in models])
            assert isinstance(map_wh_per_object, Integer[Array, "objects 2"])

            double_sided: Bool[Array, "vertices"]
            double_sided = MergedModel.generate_object_vert_info(
                counts,
                [obj.double_sided for obj in objects],
            )
            assert isinstance(double_sided, Bool[Array, "vertices"])

        # merge maps
        diffuse_map, single_map_shape = MergedModel.merge_maps(
            [m.diffuse_map for m in models])
        specular_map, _ = MergedModel.merge_maps(
            [m.specular_map for m in models])

        @jaxtyped
        @partial(jax.jit, inline=True)
        def transform_vert(
            verts: Float[Array, "N 3"],
            local_scaling: Vec3f,
            transform: ModelMatrix,
        ) -> Vertices:
            """Apply transforms defined in `ModelObject` to vertices."""
            world: Float[Array, "N 3"] = Camera.apply_pos(
                verts * local_scaling,
                transform,
            )
            assert isinstance(world, Float[Array, "N 3"])

            return world

        # merge verts
        verts, faces = MergedModel.merge_verts(
            [
                transform_vert(
                    verts=obj.model.verts,
                    local_scaling=obj.local_scaling,
                    transform=obj.transform,
                ) for obj in objects
            ],
            [m.faces for m in models],
        )

        @jaxtyped
        @partial(jax.jit, inline=True)
        def transform_normals(
            normals: Float[Array, "N 3"],
            transform: ModelMatrix,
        ) -> Vertices:
            """Apply transforms defined in `ModelObject` to vertex normals."""
            world: Float[Array, "N 3"] = Camera.apply_vec(
                normals,
                # transform by inverse transpose
                jnp.linalg.inv(transform).T,
            )
            assert isinstance(world, Float[Array, "N 3"])

            return world

        norms, faces_norm = MergedModel.merge_verts(
            [
                transform_normals(obj.model.norms, obj.transform)
                for obj in objects
            ],
            [m.faces_norm for m in models],
        )
        uvs, faces_uv = MergedModel.merge_verts(
            [m.uvs for m in models],
            [m.faces_uv for m in models],
        )

        return MergedModel(
            verts=verts,
            norms=norms,
            uvs=uvs,
            faces=faces,
            faces_norm=faces_norm,
            faces_uv=faces_uv,
            texture_shape=map_wh_per_object,
            texture_index=map_indices,
            double_sided=double_sided,
            offset=single_map_shape[0],
            diffuse_map=diffuse_map,
            specular_map=specular_map,
        )

    @staticmethod
    @jaxtyped
    @partial(jax.jit, inline=True)
    def create_camera_from_parameters(camera: CameraParameters) -> Camera:
        """Create a camera from camera parameters."""
        view_mat: View = Camera.view_matrix(
            eye=camera.position,
            centre=camera.target,
            up=camera.up,
        )
        assert isinstance(view_mat, View), f"{view_mat}"
        view_inv: View = Camera.view_matrix_inv(
            eye=camera.position,
            centre=camera.target,
            up=camera.up,
        )
        assert isinstance(view_inv, View), f"{view_inv}"
        projection_mat: Projection = Camera.perspective_projection_matrix(
            fovy=camera.vfov,
            aspect=(lax.tan(jnp.radians(camera.hfov) / 2.) /
                    lax.tan(jnp.radians(camera.vfov) / 2.)),
            z_near=camera.near,
            z_far=camera.far,
        )
        assert isinstance(projection_mat, Projection), f"{projection_mat}"
        viewport_mat: Viewport = Camera.viewport_matrix(
            lowerbound=jnp.zeros(2, dtype=int),
            dimension=jnp.array((camera.viewWidth, camera.viewHeight)),
            depth=jnp.array(camera.viewDepth),
        )
        assert isinstance(viewport_mat, Viewport), f"{viewport_mat}"

        _camera: Camera = Camera.create(
            view=view_mat,
            projection=projection_mat,
            viewport=viewport_mat,
            view_inv=view_inv,
        )
        assert isinstance(_camera, Camera), f"{_camera}"

        return _camera

    @classmethod
    @jaxtyped
    @partial(jax.jit, static_argnames=("cls", ), donate_argnums=(4, ))
    def render(
        cls,
        model: MergedModel,
        light: LightParameters,
        camera: Camera,
        buffers: Buffers,
        shadow_param: Optional[ShadowParameters] = None,
    ) -> Buffers:
        """Render the scene with the given camera.

        Parameters:
          - model: merged model of all the objects to render.
          - light: the light to render the scene with.
          - camera: the camera to render the scene with.
          - buffers: the buffers to render the scene with.
          - shadow_param: the shadow parameters to render the scene with. Keep
            it None to disable shadows.

        Returns: Buffers, with zbuffer and (coloured image, ).
        """
        # flatten so each vertex has its own "extra"
        position = model.verts[model.faces.reshape((-1, ))]
        normal = model.norms[model.faces_norm.reshape((-1, ))]
        uv = model.uvs[model.faces_uv.reshape((-1, ))]

        texture_index = model.texture_index[model.faces_uv.reshape((-1, ))]
        double_sided = model.texture_index[model.faces_uv.reshape((-1, ))]

        face_indices: Integer[Array, "_ 3"]
        face_indices = jnp.arange(model.faces.size).reshape(model.faces.shape)
        assert isinstance(face_indices, Integer[Array, "_ 3"])

        light_dir: Vec3f = normalise(light.direction.copy())
        assert isinstance(light_dir, Vec3f), f"{light_dir}"

        light_dir_eye: Vec3f = Camera.apply_vec(
            light_dir.copy(),
            camera.view,
        )
        assert isinstance(light_dir_eye, Vec3f), f"{light_dir_eye}"

        extra = PhongReflectionTextureExtraInput(
            position=position,
            normal=normal,
            uv=uv,
            light=LightSource(
                direction=light_dir,
                colour=light.colour,
            ),
            light_dir_eye=light_dir_eye,
            texture_shape=model.texture_shape,
            texture_index=texture_index,
            texture_offset=model.offset,
            texture=model.diffuse_map,
            specular_map=model.specular_map,
            ambient=light.ambient,
            diffuse=light.diffuse,
            specular=light.specular,
        )

        if shadow_param is None:
            # no shadows
            buffers = render(
                camera=camera,
                shader=PhongReflectionTextureShader,
                buffers=buffers,
                face_indices=face_indices,
                extra=extra,
            )
            assert isinstance(buffers, Buffers), f"{buffers}"

            return buffers
        else:
            # with shadows
            assert isinstance(shadow_param,
                              ShadowParameters), f"{shadow_param}"
            # first pass: render shadow map
            shadow: Shadow = Shadow.render_shadow_map(
                shadow_map=lax.full_like(
                    buffers.zbuffer,
                    DtypeInfo.create(jax.dtypes.result_type(
                        buffers.zbuffer)).max,
                ),
                verts=model.verts,
                faces=model.faces,
                light_direction=light.direction,
                viewport_matrix=camera.viewport,
                centre=shadow_param.centre,
                up=shadow_param.up,
                strength=shadow_param.strength,
                offset=shadow_param.offset,
            )
            assert isinstance(shadow, Shadow), f"{shadow}"

            _extra = PhongReflectionShadowTextureExtraInput(
                **extra._asdict(),
                shadow=shadow,
                camera=camera,
            )

            # second pass: actual rendering
            buffers = render(
                camera=camera,
                shader=PhongReflectionShadowTextureShader,
                buffers=buffers,
                face_indices=face_indices,
                extra=_extra,
            )
            assert isinstance(buffers, Buffers), f"{buffers}"

            return buffers

    @classmethod
    @jaxtyped
    def get_camera_image(
        cls,
        objects: ObjectsT,
        light: LightParameters,
        camera: Union[Camera, CameraParameters],
        width: int,
        height: int,
        colour_default: Colour = jnp.array((1., 1., 1.), dtype=jnp.single),
        zbuffer_default: Num[Array, ""] = jnp.array(1),
        zbuffer_dtype: jnp.dtype[Any] = jnp.single,
        shadow_param: Optional[ShadowParameters] = None,
    ) -> Canvas:
        """Render the scene with the given camera.

        Parameters:
          - objects: the objects to render.
          - light: the light to render the scene with.
          - camera: the camera to render the scene with.
          - width, height: the size of the image to render.
          - colourChannels: the number of colour channels to render.
          - shadow_param: the shadow parameters to render the scene with. Keep

        Returns: Buffers, with zbuffer and (coloured image, ).
        """
        _camera: Camera
        if isinstance(camera, CameraParameters):
            _camera = cls.create_camera_from_parameters(camera)
        else:
            _camera = camera

        assert isinstance(_camera, Camera), f"{_camera}"

        zbuffer: ZBuffer = lax.full(
            (width, height),
            jnp.array(zbuffer_default, dtype=zbuffer_dtype),
        )
        canvas: Canvas = jnp.full(
            (width, height, colour_default.size),
            colour_default,
        )
        buffers: Buffers = Buffers(
            zbuffer=zbuffer,
            targets=(canvas, ),
        )

        model: MergedModel = cls.merge_objects(objects)
        assert isinstance(model, MergedModel), f"{model}"

        _, (canvas, ) = cls.render(
            model=model,
            light=light,
            camera=_camera,
            buffers=buffers,
            shadow_param=shadow_param,
        )
        assert isinstance(canvas, Canvas), f"{canvas}"

        return canvas



from typing import NamedTuple

import jax
import jax.lax as lax
import jax.numpy as jnp
from jax.tree_util import Partial
from jaxtyping import Array, Bool, Float, Integer, jaxtyped

from ..geometry import (Camera, Interpolation, interpolate, normalise,
                        to_cartesian, to_homogeneous)
from ..shader import ID, MixerOutput, PerFragment, PerVertex, Shader
from ..types import (Colour, FaceIndices, LightSource, NormalMap, Texture,
                     Triangle, Vec2f, Vec2i, Vec3f, Vec4f)

jax.config.update('jax_array', True)

Triangle3f = Float[Array, "3 3"]
Triangle2f = Float[Array, "3 2"]


class PhongTextureDarbouxExtraInput(NamedTuple):
    """Extra input for Phong Shader.

    Attributes:
      - position: in world space, of each vertex.
      - normal: in world space, of each vertex.
      - uv: in texture space, of each vertex.
      - light: parallel `headlight` light source, shared by all vertices.
        It is in the eye/view space.
      - texture: texture, shared by all vertices.
      - normal_map: normal map, shared by all vertices.
        This is in Darboux frame.
      - id_to_face: id of the face that each vertex belongs to.
      - faces_indices: id of the vertex that each face contains.
    """
    position: Float[Array, "vertices 3"]  # in world space
    normal: Float[Array, "vertices 3"]  # in world space
    uv: Float[Array, "vertices 2"]  # in texture space
    light: LightSource
    texture: Texture
    normal_map: NormalMap
    """In Darboux frame."""
    id_to_face: Integer[Array, "vertices"]
    """id of the face that each vertex belongs to."""
    faces_indices: FaceIndices
    """id of the vertex that each face contains."""


class PhongTextureDarbouxExtraFragmentData(NamedTuple):
    """From vertex shader to fragment shader, and from fragment shader to mixer.

    Attributes:
      - normal: in clip space, of each fragment; From VS to FS.
      - uv: in texture space, of each fragment; From VS to FS.
      - triangle: in normalised device coordinates (NDC), of each fragment;
        From VS to FS. This should not be interpolated.
      - triangle: in texture space, of each fragment; From VS to FS.
        This should not be interpolated.
      - colour: colour when passing from FS to mixer.
    """
    normal: Vec3f = jnp.array([0.0, 0.0, 0.0])
    uv: Vec2f = jnp.zeros(2)
    triangle: Triangle3f = jnp.zeros((3, 3))
    triangle_uv: Triangle2f = jnp.zeros((3, 2))
    """triangle in NDC, not interpolated."""
    colour: Colour = jnp.array([0.0, 0.0, 0.0])


class PhongTextureDarbouxExtraMixerOutput(NamedTuple):
    """When render to only one target, for simplicity."""
    canvas: Colour


class PhongTextureDarbouxShader(Shader[PhongTextureDarbouxExtraInput,
                                       PhongTextureDarbouxExtraFragmentData,
                                       PhongTextureDarbouxExtraMixerOutput]):
    """Phong Shading with simple parallel lighting and texture, normals are
        represented in tangent space (Darboux frame)."""

    @staticmethod
    @jaxtyped
    @jax.jit
    def vertex(
        gl_VertexID: ID,
        gl_InstanceID: ID,
        camera: Camera,
        extra: PhongTextureDarbouxExtraInput,
    ) -> tuple[PerVertex, PhongTextureDarbouxExtraFragmentData]:
        # Use gl_VertexID to index in `extra` buffer.
        assert isinstance(gl_VertexID, ID), gl_VertexID
        assert isinstance(gl_InstanceID, ID), gl_InstanceID
        position: Vec4f = to_homogeneous(extra.position[gl_VertexID])
        gl_Position: Vec4f = camera.to_clip(position)
        assert isinstance(gl_Position, Vec4f)

        assert isinstance(extra.id_to_face, Integer[Array, "vertices"]), (
            f"Expected Integer array with shape {extra.position.shape[:1]}, "
            f"got {type(extra.id_to_face)} with shape "
            f"{extra.id_to_face.shape}")
        assert isinstance(extra.faces_indices, Integer[Array, "faces 3"]), (
            f"Expected Integer array with shape (faces, 3), "
            f"got {type(extra.faces_indices)} with shape "
            f"{extra.faces_indices.shape}")

        face_indices = extra.faces_indices[extra.id_to_face[gl_VertexID]]
        triangle_model: Triangle = to_homogeneous(extra.position[face_indices])
        triangle_clip: Triangle = camera.to_clip(triangle_model)
        triangle_ndc: Triangle3f = to_cartesian(triangle_clip)
        assert isinstance(triangle_ndc, Triangle3f)

        triangle_uv: Triangle2f = extra.uv[face_indices]
        assert isinstance(triangle_uv, Triangle2f)

        # assume normal here is in world space. If it is in model space, it
        # must be transformed by the inverse transpose of the model matrix.
        # Ref: https://github.com/ssloy/tinyrenderer/wiki/Lesson-5:-Moving-the-camera#transformation-of-normal-vectors
        normal: Vec3f = Camera.apply_vec(
            normalise(extra.normal[gl_VertexID]),
            camera.world_to_eye_norm,
        )
        assert isinstance(normal, Vec3f)

        return (
            PerVertex(gl_Position=gl_Position),
            PhongTextureDarbouxExtraFragmentData(
                normal=normal,
                uv=extra.uv[gl_VertexID],
                triangle=triangle_ndc,
                triangle_uv=triangle_uv,
            ),
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def interpolate(
        values: PhongTextureDarbouxExtraFragmentData,
        barycentric_screen: Vec3f,
        barycentric_clip: Vec3f,
    ) -> PhongTextureDarbouxExtraFragmentData:
        smooth_interpolation = Partial(
            interpolate,
            barycentric_screen=barycentric_screen,
            barycentric_clip=barycentric_clip,
            mode=Interpolation.SMOOTH,
        )

        normal: Vec3f = smooth_interpolation(values.normal)
        assert isinstance(normal, Vec3f)

        uv: Vec2f = smooth_interpolation(values.uv)
        assert isinstance(uv, Vec2f)

        varying: PhongTextureDarbouxExtraFragmentData = PhongTextureDarbouxExtraFragmentData(
            normal=normal,
            uv=uv,
            # pick first of the 3, as they are the same
            # noticed that `values` are batches, so here values.triangle is
            # actually in the shape of (3, 3, 3)
            triangle=values.triangle[0],
            triangle_uv=values.triangle_uv[0],
        )
        assert isinstance(varying.triangle, Triangle3f)
        assert isinstance(varying.triangle_uv, Triangle2f)

        return varying

    @staticmethod
    @jaxtyped
    @jax.jit
    def fragment(
        gl_FragCoord: Vec4f,
        gl_FrontFacing: Bool[Array, ""],
        gl_PointCoord: Vec2f,
        varying: PhongTextureDarbouxExtraFragmentData,
        extra: PhongTextureDarbouxExtraInput,
    ) -> tuple[PerFragment, PhongTextureDarbouxExtraFragmentData]:
        built_in: PerFragment = Shader.fragment(
            gl_FragCoord,
            gl_FrontFacing,
            gl_PointCoord,
            varying,
            extra,
        )[0]
        assert isinstance(built_in, PerFragment)
        assert isinstance(varying, PhongTextureDarbouxExtraFragmentData), (
            f"Expected PhongTextureDarbouxExtraFragmentData, got {varying}")

        # repeat texture
        uv = (lax.floor(varying.uv).astype(int) %
              jnp.asarray(extra.texture.shape[:2]))

        normal: Vec3f = normalise(varying.normal)
        A: Float[Array, "3 3"] = jnp.vstack([
            varying.triangle[1, :] - varying.triangle[0, :],
            varying.triangle[2, :] - varying.triangle[0, :],
            normal,
        ])
        AI: Float[Array, "3 3"] = jnp.linalg.inv(A)
        _uv: Triangle2f = varying.triangle_uv
        i: Vec3f = AI @ jnp.array(
            [_uv[1, 0] - _uv[0, 0], _uv[2, 0] - _uv[0, 0], 0])
        j: Vec3f = AI @ jnp.array(
            [_uv[1, 1] - _uv[0, 1], _uv[2, 1] - _uv[0, 1], 0])

        B: Float[Array, "3 3"] = lax.concatenate(
            [
                normalise(i)[:, None],
                normalise(j)[:, None],
                normal[:, None],
            ],
            dimension=1,
        )
        normal = normalise(B @ extra.normal_map[uv[0], uv[1]])
        assert isinstance(normal, Vec3f)

        texture_colour: Colour = extra.texture[uv[0], uv[1]]

        # light colour * intensity
        light_colour: Colour = extra.light.colour * lax.dot(
            normal,
            normalise(extra.light.direction),
        )

        return (
            PerFragment(
                keeps=jnp.logical_and(built_in.keeps, gl_FrontFacing),
                use_default_depth=built_in.use_default_depth,
            ),
            varying._replace(colour=lax.cond(
                (light_colour >= 0).all(),
                lambda: texture_colour * light_colour,
                lambda: jnp.zeros(3),
            )),
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def mix(
        gl_FragDepth: Float[Array, "primitives"],
        keeps: Bool[Array, "primitives"],
        extra: PhongTextureDarbouxExtraFragmentData,
    ) -> tuple[MixerOutput, PhongTextureDarbouxExtraMixerOutput]:
        mixer_output: MixerOutput
        extra_output: PhongTextureDarbouxExtraFragmentData
        mixer_output, extra_output = Shader.mix(gl_FragDepth, keeps, extra)
        assert isinstance(mixer_output, MixerOutput)
        assert isinstance(extra_output, PhongTextureDarbouxExtraFragmentData)

        return (
            mixer_output,
            PhongTextureDarbouxExtraMixerOutput(canvas=extra_output.colour),
        )



from typing import NamedTuple, Sequence, Union

import jax
import jax.experimental.checkify as checkify
import jax.lax as lax
import jax.numpy as jnp
from jax.tree_util import tree_map
from jaxtyping import Array, Bool, Float, Integer, Num, Shaped, PyTree, jaxtyped

from .types import (FALSE_ARRAY, FaceIndices, Normals, SpecularMap, Texture,
                    UVCoordinates, Vec3f, Vertices)
from .value_checker import index_in_bound

ModelMatrix = Float[Array, "4 4"]


class Model(NamedTuple):
    """Model with vertices specification and attached maps.

    NormalMap is not included for now as it is not used in the reference
    implementation
    [erwincoumans/tinyrenderer](https://github.com/erwincoumans/tinyrenderer).
    """
    verts: Vertices
    norms: Normals
    uvs: UVCoordinates
    faces: FaceIndices
    faces_norm: FaceIndices
    faces_uv: FaceIndices

    diffuse_map: Texture
    specular_map: SpecularMap

    @jaxtyped
    @jax.jit
    def asserts(self):
        """Asserts that all fields are of correct shape and type."""
        assert isinstance(self.verts, Vertices), f"{self.verts}"
        assert isinstance(self.norms, Normals), f"{self.norms}"
        assert isinstance(self.uvs, UVCoordinates), f"{self.uvs}"
        assert isinstance(self.faces, FaceIndices), f"{self.faces}"
        assert isinstance(self.faces_norm, FaceIndices), f"{self.faces_norm}"
        assert isinstance(self.faces_uv, FaceIndices), f"{self.faces_uv}"
        assert isinstance(self.diffuse_map, Texture), f"{self.diffuse_map}"
        assert isinstance(self.specular_map,
                          SpecularMap), f"{self.specular_map}"

    @jaxtyped
    @checkify.checkify
    @jax.jit
    def value_checks(self):
        """Check values are of correct value ranges.

        Checks implemented:
          - indices are in bound, for faces, faces_norm, faces_uv (against
            verts, norms, uv).

        UV coordinates are not checked as out-of-bound values are allowed and used as "repeat" mode.

        Usage:
          model = Model(...)
          err, _ = model.value_checks()
          err.throw()  # throw if any error is found.
        """
        # indices are in bound
        checkify.check(
            index_in_bound(self.faces, self.verts.shape[0]),
            # f-string for shape as shape is compile-time constant and
            # is not supported as a format string parameter in checkify.
            f"faces out of bound, expected [0, {self.verts.shape[0]}),"
            # separately specify fmt_kwargs in a non-f-string.
            " got {max_idx}.",
            max_idx=self.faces.max(),
        )
        checkify.check(
            index_in_bound(self.faces_norm, self.norms.shape[0]),
            f"faces_norm out of bound, expected [0, {self.norms.shape[0]}),"
            " got {max_idx}.",
            max_idx=self.faces_norm.max(),
        )
        checkify.check(
            index_in_bound(self.faces_uv, self.uvs.shape[0]),
            f"faces_uv out of bound, expected [0, {self.uvs.shape[0]}),"
            " got {max_idx}.",
            max_idx=self.faces_uv.max(),
        )


VertT = Num[Array, "_dim ..."]
VertsT = Union[list[VertT], tuple[VertT, ...]]
FaceIndicesT = Num[Array, "_faces 3"]
FaceIndicessT = Union[list[FaceIndicesT], tuple[FaceIndicesT, ...]]

MapT = Num[Array, "_width _height ..."]
MapsT = Union[list[MapT], tuple[MapT, ...]]
Shape2DT = tuple[Integer[Array, ""], Integer[Array, ""]]
"""Shape of first two components of a map, i.e., (width, height)."""


class MergedModel(NamedTuple):
    """Merged model with vertices, normals, uv coordinates, and faces."""
    verts: Vertices
    norms: Normals
    uvs: UVCoordinates
    faces: FaceIndices
    faces_norm: FaceIndices
    faces_uv: FaceIndices

    # broadcasted object info into per-vertex
    texture_index: Integer[Array, "vertices"]
    """Texture map index for each vertex."""
    double_sided: Bool[Array, "vertices"]
    """Whether each face is double sided."""

    # Merged maps
    texture_shape: Integer[Array, "objects 2"]
    """Width, height of each texture map."""
    offset: Integer[Array, ""]
    """Width of biggest merged maps, as [0] returned by `merge_maps`."""
    diffuse_map: Texture
    specular_map: SpecularMap

    @staticmethod
    @jaxtyped
    def generate_object_vert_info(
        counts: Sequence[int],
        values: Sequence[Shaped[Array, "..."]],
    ) -> Shaped[Array, "vertices"]:
        """Generate object-wide info for each vertex in merged model as
            vertex-level info.

        Parameters:
          - counts: Number of vertices of each object.
          - values: value to be filled in for each object.

        Returns: Map indices for each face.

        Note: this function cannot be jitted by itself as it uses the value of
            `counts` to create matrices: the shape depends on the value of
            `counts`.
        """
        values: Sequence[Shaped[Array, "_"]] = tree_map(
            lambda count, value: jnp.full(
                (count, *jnp.asarray(value).shape), value),
            counts,
            values,
        )
        map_indices = lax.concatenate(values, dimension=0)

        return map_indices

    @staticmethod
    @jaxtyped
    @jax.jit
    def merge_verts(
        vs: VertsT,
        fs: FaceIndicessT,
    ) -> tuple[VertT, FaceIndicesT]:
        """"""
        counts = [v.shape[0] for v in vs[:-1]]
        cumsum = [0]
        for count in counts:
            cumsum.append(cumsum[-1] + count)

        dtype = jax.dtypes.result_type(*vs)

        # merge vertices
        verts: VertT = lax.concatenate(
            [v.astype(dtype) for v in vs],
            dimension=0,
        )
        # merge faces
        faces: FaceIndicesT = lax.concatenate(
            [f + cumsum[i] for i, f in enumerate(fs)],
            dimension=0,
        )

        return verts, faces

    @staticmethod
    @jaxtyped
    @jax.jit
    def merge_maps(maps: MapsT) -> tuple[MapT, Shape2DT]:
        """Merge maps by concatenating them along the first axis.

        All maps must have the same number of dimensions, i.e., `len(m.shape)`
        must be the same for all maps `m`.

        If the other axises is not the same, they are padded with undefined
        values, keeping [0:shape[i]] as given values, and [shape[i]:] being
        padded values at dimension i. If they are of different dtypes, they are
        promoted together.

        Parameters:
          - maps: a list of maps to merge.

        Returns:
          - A merged map.
          - shape of first and second axis of each map.
        """
        # TODO: find a better way to merge maps
        dims: int = len(maps[0].shape)
        shapes: PyTree[tuple[Integer[Array, ""], ...], ...]
        shapes = tree_map(lambda m: m.shape, maps)
        # pick the largest shape for each dimension
        single_shape: tuple[Integer[Array, ""], ...]
        single_shape = tuple(
            (max((shape[i] for shape in shapes)) for i in range(dims)))

        dtype = jax.dtypes.result_type(*maps)

        new_map = lax.concatenate(
            tree_map(
                lambda m: lax.pad(
                    m,
                    jnp.array(0, dtype=m.dtype),
                    tree_map(
                        lambda capacity, content: (0, capacity - content, 0),
                        single_shape,
                        m.shape,
                    ),
                ).astype(dtype),
                maps,
            ),
            dimension=0,
        )

        return new_map, single_shape[:2]

    @staticmethod
    @jaxtyped
    @jax.jit
    def uv_repeat(
        uv: Float[Array, "2"],
        shape: Integer[Array, "2"],
        map_index: Integer[Array, ""],
        offset: Integer[Array, ""],
    ) -> Float[Array, "2"]:
        """Compute final UV coordinates as if it is repeatedly tiled (in case
            of out-of-bound uv coordinate).

        Parameters:
          - uv: raw uv coordinates, in floating numbers. Only fractional part
            is used, as if the uv coordinates are in [0, 1].
          - shape: of the map being used, according to `map_index`.
          - offset: of each map in the merged maps, as [0] returned by
            `merge_maps`. Only first axis is required, thus here we just need a
            scalar.
          - map_index: index of the map to use.
        """
        assert isinstance(uv, Float[Array, "2"]), f"{uv}"
        assert isinstance(shape, Integer[Array, "2"]), f"{shape}"
        assert isinstance(map_index, Integer[Array, ""]), f"{map_index}"
        assert isinstance(offset, Integer[Array, ""]), f"{offset}"
        # since given uv are in [0, 1] (and may be scaled, if is cube),
        # we need to multiply it by (w, h) of the texture map first.
        # This is equivalent to just obtain the fractional part of uv.
        fractional_uv, _ = jnp.modf(uv)
        fractional_uv = jnp.where(
            fractional_uv < 0,
            fractional_uv + 1,
            fractional_uv,
        )
        assert isinstance(fractional_uv, Float[Array, "2"])

        return (fractional_uv * shape).at[0].add(map_index * offset)


class ModelObject(NamedTuple):
    """Model object with model and transform."""
    model: Model
    """Reference to the model, with mesh, attached maps, etc."""
    local_scaling: Vec3f = jnp.ones(3)
    """Local scaling factors of the object, in x, y, z."""
    transform: ModelMatrix = jnp.identity(4)
    """Transform matrix (model matrix) of the model."""
    # TODO: Support double_sided
    double_sided: Bool[Array, ""] = FALSE_ARRAY
    """Whether the object is double-sided."""



from typing import NamedTuple

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, jaxtyped

from ..geometry import Camera, normalise, to_homogeneous
from ..shader import ID, MixerOutput, PerFragment, PerVertex, Shader
from ..types import Colour, LightSource, Texture, Vec2f, Vec3f, Vec4f

jax.config.update('jax_array', True)


class PhongTextureExtraInput(NamedTuple):
    """Extra input for Phong Shader.

    Attributes:
      - position: in world space, of each vertex.
      - normal: in world space, of each vertex.
      - uv: in texture space, of each vertex.
      - light: parallel `headlight` light source, shared by all vertices.
        It is in the eye/view space.
      - texture: texture, shared by all vertices.
    """
    position: Float[Array, "vertices 3"]  # in world space
    normal: Float[Array, "vertices 3"]  # in world space
    uv: Float[Array, "vertices 2"]  # in texture space
    light: LightSource
    texture: Texture


class PhongTextureExtraFragmentData(NamedTuple):
    """From vertex shader to fragment shader, and from fragment shader to mixer.

    Attributes:
      - normal: in clip space, of each fragment; From VS to FS.
      - uv: in texture space, of each fragment; From VS to FS.
      - colour: colour when passing from FS to mixer.
    """
    normal: Vec3f = jnp.array([0.0, 0.0, 0.0])
    uv: Vec2f = jnp.zeros(2)
    colour: Colour = jnp.array([0.0, 0.0, 0.0])


class PhongTextureExtraMixerOutput(NamedTuple):
    """When render to only one target, for simplicity."""
    canvas: Colour


class PhongTextureShader(Shader[PhongTextureExtraInput,
                                PhongTextureExtraFragmentData,
                                PhongTextureExtraMixerOutput]):
    """Phong Shading with simple parallel lighting and texture."""

    @staticmethod
    @jaxtyped
    @jax.jit
    def vertex(
        gl_VertexID: ID,
        gl_InstanceID: ID,
        camera: Camera,
        extra: PhongTextureExtraInput,
    ) -> tuple[PerVertex, PhongTextureExtraFragmentData]:
        # Use gl_VertexID to index in `extra` buffer.
        position: Vec4f = to_homogeneous(extra.position[gl_VertexID])
        gl_Position: Vec4f = camera.to_clip(position)
        assert isinstance(gl_Position, Vec4f)

        # assume normal here is in world space. If it is in model space, it
        # must be transformed by the inverse transpose of the model matrix.
        # Ref: https://github.com/ssloy/tinyrenderer/wiki/Lesson-5:-Moving-the-camera#transformation-of-normal-vectors
        normal: Vec3f = Camera.apply_vec(
            normalise(extra.normal[gl_VertexID]),
            camera.world_to_eye_norm,
        )
        assert isinstance(normal, Vec3f)

        return (
            PerVertex(gl_Position=gl_Position),
            PhongTextureExtraFragmentData(
                normal=normal,
                uv=extra.uv[gl_VertexID],
            ),
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def fragment(
        gl_FragCoord: Vec4f,
        gl_FrontFacing: Bool[Array, ""],
        gl_PointCoord: Vec2f,
        varying: PhongTextureExtraFragmentData,
        extra: PhongTextureExtraInput,
    ) -> tuple[PerFragment, PhongTextureExtraFragmentData]:
        built_in: PerFragment = Shader.fragment(
            gl_FragCoord,
            gl_FrontFacing,
            gl_PointCoord,
            varying,
            extra,
        )[0]
        assert isinstance(built_in, PerFragment)

        # repeat texture
        uv = (lax.floor(varying.uv).astype(int) %
              jnp.asarray(extra.texture.shape[:2]))
        texture_colour: Colour = extra.texture[uv[0], uv[1]]

        # light colour * intensity
        light_colour: Colour = extra.light.colour * lax.dot(
            normalise(varying.normal),
            normalise(extra.light.direction),
        )

        return (
            PerFragment(
                keeps=jnp.logical_and(built_in.keeps, gl_FrontFacing),
                use_default_depth=built_in.use_default_depth,
            ),
            PhongTextureExtraFragmentData(
                colour=lax.cond(
                    (light_colour >= 0).all(),
                    lambda: texture_colour * light_colour,
                    lambda: jnp.zeros(3),
                ),
                uv=varying.uv,
                normal=varying.normal,
            ),
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def mix(
        gl_FragDepth: Float[Array, "primitives"],
        keeps: Bool[Array, "primitives"],
        extra: PhongTextureExtraFragmentData,
    ) -> tuple[MixerOutput, PhongTextureExtraMixerOutput]:
        mixer_output: MixerOutput
        extra_output: PhongTextureExtraFragmentData
        mixer_output, extra_output = Shader.mix(gl_FragDepth, keeps, extra)
        assert isinstance(mixer_output, MixerOutput)
        assert isinstance(extra_output, PhongTextureExtraFragmentData)

        return (
            mixer_output,
            PhongTextureExtraMixerOutput(canvas=extra_output.colour),
        )



from typing import NamedTuple, NewType, Union

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, jaxtyped

from .geometry import transform_matrix_from_rotation
from .model import Model, ModelObject
from .shapes.capsule import UpAxis, create_capsule
from .shapes.cube import create_cube
from .types import SpecularMap, Texture, Vec3f, Vec4f

GUID = NewType("GUID", int)


class Scene(NamedTuple):
    """Scene with models and objects. Noticed that with each update to the
        scene, the scene instance is replaced with a new one.
    """
    guid: GUID = GUID(0)
    """Max unique identifier among all objects in the scene. It equals to the
        numbers of models and objects ever created in the scene.
    """
    models: dict[GUID, Model] = {}
    """Models in the scene, indexed by their unique identifier."""
    objects: dict[GUID, ModelObject] = {}
    """Objects in the scene, indexed by their unique identifier."""

    @jaxtyped
    def add_model(self, model: Model) -> tuple["Scene", GUID]:
        """Add a model to the scene.

        Parameters:
          - model: a model to add to the scene.

        Returns:
          A tuple of the updated scene and the unique identifier of the model.
        """
        guid = self.guid
        new_scene = self._replace(
            guid=GUID(guid + 1),
            models={
                **self.models, guid: model
            },
        )

        return new_scene, guid

    @jaxtyped
    def add_cube(
        self,
        half_extents: Union[Float[Array, "3"], tuple[float, float, float]],
        diffuse_map: Texture,
        texture_scaling: Union[Float[Array, "2"], tuple[float, float]],
    ) -> tuple["Scene", GUID]:
        """Add a cube to the scene.

        Parameters:
          - half_extents: the half-size of the cube. The final cube would have
            x-y-z dimension of 2 * half_extents.
          - diffuse_map: the diffuse map of the cube.
          - texture_scaling: the scaling factor of the texture, in x and y.
        """
        # reference: https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/model.cpp#L215
        specular_map: SpecularMap = lax.full(diffuse_map.shape[:2], 2.0)

        _half_extents = jnp.asarray(half_extents)
        assert isinstance(_half_extents, Float[Array, "3"]), (
            "Expected 2 floats in half_extends, got {half_extents}")

        _texture_scaling = jnp.asarray(texture_scaling)
        assert isinstance(_texture_scaling, Float[Array, "2"]), (
            "Expected 2 floats in texture_scaling, got {texture_scaling}")

        model: Model = create_cube(
            half_extents=_half_extents,
            texture_scaling=_texture_scaling,
            diffuse_map=diffuse_map,
            specular_map=specular_map,
        )

        return self.add_model(model)

    @jaxtyped
    def add_capsule(
        self,
        radius: float,
        half_height: float,
        up_axis: UpAxis,
        diffuse_map: Texture,
    ) -> tuple["Scene", GUID]:
        """Add a capsule to the scene.

        Parameters:
          - radius: the radius of the capsule.
          - half_height: the half height of the capsule.
          - up_axis: the up axis of the capsule.
          - diffuse_map: the diffuse map of the capsule.
        """
        # reference: https://github.com/erwincoumans/tinyrenderer/blob/89e8adafb35ecf5134e7b17b71b0f825939dc6d9/model.cpp#L215
        specular_map: SpecularMap = lax.full(diffuse_map.shape[:2], 2.0)
        model: Model = create_capsule(
            radius=jnp.asarray(radius),
            half_height=jnp.asarray(half_height),
            up_axis=up_axis,
            diffuse_map=diffuse_map,
            specular_map=specular_map,
        )

        return self.add_model(model)

    @jaxtyped
    def add_object_instance(self, model_id: GUID) -> tuple["Scene", GUID]:
        """Add an object instance to the scene.

        Parameters:
          - model_id: the unique identifier of the model to add.

        Returns:
          A tuple of the updated scene and the unique identifier of the object.
        """
        guid = self.guid
        new_scene = self._replace(
            guid=GUID(guid + 1),
            objects={
                **self.objects, guid: ModelObject(model=self.models[model_id])
            },
        )

        return new_scene, guid

    @jaxtyped
    def delete_model(self, model_id: GUID, check: bool) -> "Scene":
        """Delete a model from the scene.

        Parameters:
          - model_id: the unique identifier of the model to delete.
          - check: whether to check if the model exists and/or being used.
        """
        if model_id not in self.models:
            if check:
                raise RuntimeError(f"model {model_id} does not exist")

            return self

        model = self.models[model_id]
        if check:
            for object_id, object in self.objects.items():
                if object.model == model:
                    raise RuntimeError(
                        f"model {model_id} is being used by object"
                        f" {object_id}")

        models = {k: v for k, v in self.models.items() if k != model_id}
        del model

        return self._replace(models=models)

    @jaxtyped
    def delete_object(self, object_id: GUID, check: bool) -> "Scene":
        """Delete an object from the scene.

        Parameters:
          - object_id: the unique identifier of the object to delete.
          - check: whether to check if the object exists.
        """
        if object_id not in self.objects:
            if check:
                raise RuntimeError(f"object {object_id} does not exist")

            return self

        _object = self.objects[object_id]
        objects = {k: v for k, v in self.objects.items() if k != object_id}
        del _object

        return self._replace(objects=objects)

    @jaxtyped
    def set_object_position(
        self,
        object_id: GUID,
        position: Union[Vec3f, tuple[float, float, float]],
    ) -> "Scene":
        """Set the position of an object in the scene.

        Parameters:
          - object_id: the unique identifier of the object.
          - position: the new position of the object.
        """
        position = jnp.asarray(position, dtype=float)
        assert isinstance(position, Vec3f), f"{position}"

        obj: ModelObject = self.objects[object_id]
        new_mat: Float[Array, "4 4"] = obj.transform.at[:3, 3].set(position)
        new_obj: ModelObject = obj._replace(transform=new_mat)

        return self._replace(objects=self.objects | {object_id: new_obj})

    @jaxtyped
    def set_object_orientation(
        self,
        object_id: GUID,
        orientation: Vec4f,
    ) -> "Scene":
        """Set the orientation of an object in the scene.

        Parameters:
          - object_id: the unique identifier of the object.
          - orientation: the new orientation of the object.
        """
        mat: Float[Array, "3 3"] = transform_matrix_from_rotation(orientation)
        assert isinstance(mat, Float[Array, "3 3"]), f"{mat}"

        obj: ModelObject = self.objects[object_id]
        new_mat: Float[Array, "4 4"] = obj.transform.at[:3, :3].set(mat)
        new_obj: ModelObject = obj._replace(transform=new_mat)

        return self._replace(objects=self.objects | {object_id: new_obj})

    @jaxtyped
    def set_object_local_scaling(
        self,
        object_id: GUID,
        local_scaling: Vec3f,
    ) -> "Scene":
        """Set the local scaling of an object in the scene.

        Parameters:
          - object_id: the unique identifier of the object.
          - local_scaling: the new local scaling of the object.
        """
        obj: ModelObject = self.objects[object_id]
        new_obj: ModelObject = obj._replace(local_scaling=local_scaling)

        return self._replace(objects=self.objects | {object_id: new_obj})

    @jaxtyped
    def set_object_double_sided(
        self,
        object_id: GUID,
        double_sided: Union[bool, Bool[Array, ""]],
    ) -> "Scene":
        """Set whether an object in the scene is double-sided.

        Parameters:
          - object_id: the unique identifier of the object.
          - double_sided: whether the object is double-sided.
        """
        obj: ModelObject = self.objects[object_id]
        new_obj: ModelObject
        new_obj = obj._replace(double_sided=jnp.asarray(double_sided))

        return self._replace(objects=self.objects | {object_id: new_obj})



from abc import ABC, abstractmethod
from typing import Generic, NamedTuple, TypeVar

import jax
import jax.lax as lax
import jax.numpy as jnp
from jax.tree_util import Partial, tree_map
from jaxtyping import Array, Bool, Float, Integer, PyTree, Shaped, jaxtyped

from .geometry import Camera, Interpolation, interpolate
from .types import FALSE_ARRAY, INF_ARRAY, TRUE_ARRAY, Vec2f, Vec3f, Vec4f

jax.config.update('jax_array', True)

ID = Integer[Array, ""]

ShaderExtraInputT = TypeVar(
    'ShaderExtraInputT',
    bound=PyTree[Shaped[Array, "..."]],
)
"""Extra input for vertex shader & fragment shader, shared by all."""


class PerVertex(NamedTuple):
    """Built-in output from Vertex Shader.

    gl_Position is in clip-space.
    """
    gl_Position: Vec4f
    # gl_PointSize is meaningful only when rendering point primitives.
    # not supported for now
    # gl_PointSize: Float[Array, ""]


class PerFragment(NamedTuple):
    """Built-in Output from Fragment Shader.

    If use_default_depth is True (default False), gl_FragCoord[2] will be used
    later by default.
    """
    gl_FragDepth: Float[Array, ""] = INF_ARRAY
    # not discard
    keeps: Bool[Array, ""] = TRUE_ARRAY
    use_default_depth: Bool[Array, ""] = FALSE_ARRAY


VaryingT = TypeVar(
    "VaryingT",
    bound=tuple[Shaped[Array, "..."], ...],
)
"""The user-defined input and second (extra) output of fragment shader."""

MixedExtraT = TypeVar(
    "MixedExtraT",
    bound=tuple[Shaped[Array, "..."], ...],
)
"""The user-defined second (extra) output of mix shader."""


class MixerOutput(NamedTuple):
    """Built-in output from `Shader.mix`.

    keep: bool, whether the output should be used to update buffers
    zbuffer: store depth value, and the result is used to set zbuffer.
    """
    keep: Bool[Array, ""]
    zbuffer: Float[Array, ""]


class Shader(ABC, Generic[ShaderExtraInputT, VaryingT, MixedExtraT]):
    """Base class for customised shader.

    Since JAX is pure functional (stateless), the state will be passed by
    returned values (the second return value in each function) in each process.
    """

    @staticmethod
    @jaxtyped
    @jax.jit
    @abstractmethod
    def vertex(
        gl_VertexID: ID,
        gl_InstanceID: ID,
        camera: Camera,
        extra: ShaderExtraInputT,
    ) -> tuple[PerVertex, VaryingT]:
        """Override this to implement the vertex shader as defined by OpenGL.

        The meaning of the inputs follows the definitions in OpenGL. Additional
        named parameters can be defined and passed in if you need, as defined
        in `extra`.

        Noticed that no internal state will be tracked, thus if there is any
        value to be passed to downstream process, it must be returned as the
        output `VaryingT` of this function.

        Relevant information from the original document is copied below.

        Parameters:
          - gl_VertexID
            the index of the vertex currently being processed. When using
            non-indexed rendering, it is the effective index of the current
            vertex (the number of vertices processed + the `first` value). For
            indexed rendering, it is the index used to fetch this vertex from
            the buffer.
          - gl_InstanceID
            the index of the current instance when doing some form of instanced
            rendering. The instance count always starts at 0, even when using
            base instance calls. When not using instanced rendering, this value
            will be 0.
          - camera: Camera [extra input, not in GLSL]
            contains view, viewport, and projection matrices.
          - extra: Camera [extra input, not in GLSL]
            User-defined extra input for vertex shader, shared by all. They are
            **not** split over batch axis 0, if any; but directly passed in.


        Return: PerVertex (used for internals) and ExtraPerVertexOutput to be
            interpolated and used by downstream pipelines.
          - gl_Position
            the clip-space output position of the current vertex.
          - gl_PointSize
            the pixel width/height of the point being rasterized. It only has a
            meaning when rendering point primitives. It will be clamped to the
            GL_POINT_SIZE_RANGE.

        Reference:
          - [Vertex Shader/Defined Inputs](https://www.khronos.org/opengl/wiki/Vertex_Shader/Defined_Inputs)
          - [Vertex Shader#Outputs](https://www.khronos.org/opengl/wiki/Vertex_Shader#Outputs)
        """
        raise NotImplementedError("vertex shader not implemented")

    @staticmethod
    @jaxtyped
    @jax.jit
    def interpolate(
        values: VaryingT,
        barycentric_screen: Vec3f,
        barycentric_clip: Vec3f,
    ) -> VaryingT:
        """Override this to customise the interpolation of user-defined inputs.

        The default implementation is to interpolate all the fields of given
        values as `smooth`, which is perspective interpolation, as defined in
        GLSL.

        Parameters:
          - values: values at the vertices of the triangle, with axis 0 being
            the batch axis. It is expected to be a tuple of multiple batched
            values.
          - barycentric_screen: barycentric coordinates in screen space of the
            point to interpolate
          - barycentric_clip: barycentric coordinates in clip space of the
            point to interpolate

        Return: interpolated values for fragment shader process, with same
            structure (order of members) as `values`
        """
        varying: VaryingT = tree_map(
            Partial(
                interpolate,
                barycentric_screen=barycentric_screen,
                barycentric_clip=barycentric_clip,
                mode=Interpolation.SMOOTH,
            ),
            values,
        )

        return varying

    @staticmethod
    @jaxtyped
    @jax.jit
    def fragment(
        gl_FragCoord: Vec4f,
        gl_FrontFacing: Bool[Array, ""],
        gl_PointCoord: Vec2f,
        varying: VaryingT,
        extra: ShaderExtraInputT,
    ) -> tuple[PerFragment, VaryingT]:
        """Override this to implement the vertex shader as defined by OpenGL.

        This is optional. The default implementation writes nothing and thus
        `gl_FragDepth` further down the pipeline will use `gl_FragCoord[2]`.
        For the `varying` input, it will be returned directly untouched.

        If the output from this default implementation is re-used, noticed that
        `use_default_depth` needs to be updated to False, otherwise the default
        depth (`gl_FragCoord[2]`) will be used in further process.

        Parameters:
          - gl_FragCoord: homogeneous coordinates in screen device space.
          - gl_FrontFacing: True if the primitive is NOT back facing.
          - gl_PointCoord: 2d coordinates in screen device space.
          - varying: interpolated values from `Shader.interpolate`; these are
            generated from `Shader.vertex`.
          - extra: ShaderExtraInputT, same as `extra` in `Shader.vertex`.

        Return: PerFragment for depth test and further mixing process.
          - gl_FragDepth: if not set (remains None), gl_FragCoord[2] will be
            used later.
          - varying: defined by user, passed as `varying` in `Shader.mix`.
            **NOTE** the return type must be the same type as `values` in
            `Shader.interpolate`, as that will be used as the dummy value for
            this return value, when `PerFragment` suggests `keeps` is False.

        Reference:
          - [Fragment Shader/Defined Inputs](https://www.khronos.org/opengl/wiki/Fragment_Shader/Defined_Inputs)
          - [Fragment Shader#Outputs](https://www.khronos.org/opengl/wiki/Fragment_Shader#Outputs)
        """
        return PerFragment(use_default_depth=TRUE_ARRAY), varying

    @staticmethod
    @jaxtyped
    @jax.jit
    def mix(
        gl_FragDepth: Float[Array, "primitives"],
        keeps: Bool[Array, "primitives"],
        extra: VaryingT,
    ) -> tuple[MixerOutput, MixedExtraT]:
        """Override this to customise the mixing behaviour per fragment over
            different primitives (triangles).

        Use this to implement the `blending` behaviour if needed.

        For the default behaviour, the values from fragment with maximum
        `gl_FragDepth` value AND `keeps` being True will be used as the output.
        In the default implementation, if no fragment has `keeps` being True,
        then mixed value will be the first fragment's value for both
        `gl_FragDepth` and `extra`.

        Returns: Built-in MixerOutput and user-defined extras.
          - MixerOutput:
            - keep: bool, whether uses this value to set the corresponding
              pixel in the buffers
            - zbuffer, the value used to update the zbuffer
          - User-defined outputs, must be a tuple (can be NamedTuple)
            Each field must be defined as same order as in the class `Buffers`.
            The values will be directly set to the `Buffers` **in the same
            order of the given definition** as if a usual `tuple`, but not
            based on field name.

        Reference:
          - [Blending](https://www.khronos.org/opengl/wiki/Blending)
        """

        def has_kept_fragment() -> Integer[Array, ""]:
            depths: Float[Array, "primitives"]
            depths = jnp.where(keeps, gl_FragDepth, jnp.inf)
            assert isinstance(depths, Float[Array, "primitives"])

            idx: Integer[Array, ""] = jnp.argmin(depths)

            return idx

        has_valid_fragment = jnp.any(keeps)

        idx: Integer[Array, ""] = lax.cond(
            has_valid_fragment,
            has_kept_fragment,
            lambda: jnp.array(0),
        )
        depth: Float[Array, ""] = gl_FragDepth[idx]
        assert isinstance(depth, Float[Array, ""])

        return (
            MixerOutput(keep=has_valid_fragment, zbuffer=depth),
            tree_map(lambda x: x[idx], extra),
        )



from typing import NamedTuple

import jax
from jaxtyping import Array, Float, jaxtyped

from ..shader import ID, PerVertex, Shader
from ..geometry import Camera, to_homogeneous
from ..types import Vec4f

jax.config.update('jax_array', True)


class DepthExtraInput(NamedTuple):
    """Extra input for Depth Shader.

    Attributes:
      - position: in world space, of each vertex.
    """
    position: Float[Array, "vertices 3"]  # in world space


class DepthExtraFragmentData(NamedTuple):
    pass


class DepthExtraMixerOutput(NamedTuple):
    pass


class DepthShader(Shader[DepthExtraInput, DepthExtraFragmentData,
                         DepthExtraMixerOutput]):
    """Depth Shading."""

    @staticmethod
    @jaxtyped
    @jax.jit
    def vertex(
        gl_VertexID: ID,
        gl_InstanceID: ID,
        camera: Camera,
        extra: DepthExtraInput,
    ) -> tuple[PerVertex, DepthExtraFragmentData]:
        # Use gl_VertexID to index in `extra` buffer.
        position: Vec4f = to_homogeneous(extra.position[gl_VertexID])
        gl_Position: Vec4f = camera.to_clip(position)
        assert isinstance(gl_Position, Vec4f)

        return (
            PerVertex(gl_Position=gl_Position),
            DepthExtraFragmentData(),
        )



from typing import Sequence, Union

import jax
import jax.numpy as jnp
from jax import lax
from jaxtyping import Array, Integer, Num, Shaped, jaxtyped

from .types import Canvas, Texture, ZBuffer


@jaxtyped
@jax.jit
def get_value_from_index(
    matrix: Shaped[Array, "width height batch *valueDimensions"],
    index: Integer[Array, "width height"],
) -> Shaped[Array, "width height *valueDimensions"]:
    """Retrieve value along 3rd axis using index value from index matrix."""
    return jax.vmap(jax.vmap(lambda mt, ix: mt[ix]))(matrix, index)


@jaxtyped
@jax.jit
def merge_canvases(
    zbuffers: Num[Array, "batch width height"],
    canvases: Shaped[Array, "batch width height channel"],
) -> tuple[ZBuffer, Canvas]:
    """Merge canvases by selecting each pixel with max z value in zbuffer,
        then merge zbuffer as well.
    """
    pixel_idx: Integer[Array, "width height"] = jnp.argmax(zbuffers, axis=0)
    assert isinstance(pixel_idx, Integer[Array, "width height"])

    zbuffer: ZBuffer = get_value_from_index(
        lax.transpose(zbuffers, (1, 2, 0)),
        pixel_idx,
    )
    assert isinstance(zbuffer, ZBuffer)

    canvas: Canvas = get_value_from_index(
        # first vmap along width, then height, then choose among "faces"
        lax.transpose(canvases, (1, 2, 0, 3)),
        pixel_idx,
    )
    assert isinstance(canvas, Canvas)

    return zbuffer, canvas


@jaxtyped
@jax.jit
def transpose_for_display(
        matrix: Num[Array,
                    "fst snd *channel"]) -> Num[Array, "snd fst *channel"]:
    return jnp.swapaxes(matrix, 0, 1)


@jaxtyped
def build_texture_from_PyTinyrenderer(
    texture: Union[Num[Array, "length"], Sequence[float]],
    width: int,
    height: int,
) -> Texture:
    """Build a texture from PyTinyrenderer's format.

    The texture was specified in C order (channel varies the fastest), but with
    y as the first axis. Besides, after swapping the first two axes, the second axis is reversed as required by this renderer.

    Parameters:
      - texture: a 1D array of length `width * height * channels`, where each
        channel elements represent a pixel in RGB order. When channels is 1,
        the resulted texture still has 3 dimensions, with last dimension of
        side 1.
      - width: width of the texture.
      - height: height of the texture.

    Returns: A texture with shape `(width, height, channels)`.
    """
    return jnp.reshape(texture, (width, height, -1),
                       order="C").swapaxes(0, 1)[:, ::-1, :]



from typing import NamedTuple

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Float, Integer, jaxtyped

from ..geometry import Camera, normalise, normalise_homogeneous, to_homogeneous
from ..model import MergedModel
from ..shader import ID, MixerOutput, PerFragment, PerVertex, Shader
from ..shadow import Shadow
from ..types import (Colour, LightSource, SpecularMap, Texture, Vec2f, Vec2i,
                     Vec3f, Vec4f)

jax.config.update('jax_array', True)


class PhongReflectionShadowTextureExtraInput(NamedTuple):
    """Extra input for Phong Reflection with Shadow Shader.

    Attributes:
      - position: in world space, of each vertex.
      - normal: in world space, of each vertex.
      - uv: in texture space, of each vertex.
      - light: parallel light source, shared by all vertices.
      - light_dir_eye: normalised light source direction in eye/view space.
      - texture_shape: shape of each texture map, shared by all vertices.
      - texture_index: index of texture map for each vertex.
      - offset_shape: offset of each texture map in the merged model.
      - texture: texture, shared by all vertices.
      - specular_map: specular map, shared by all vertices.
      - shadow: Shadow from first pass, shared by all vertices.
      - shadow_mat: shadow matrix from target screen space to shadow's screen
        space, shared by all vertices.
      - ambient: ambient strength, shared by all vertices.
      - diffuse: diffuse strength, shared by all vertices.
      - specular: specular strength, shared by all vertices.
    """
    position: Float[Array, "vertices 3"]  # in world space
    normal: Float[Array, "vertices 3"]  # in world space
    uv: Float[Array, "vertices 2"]  # in texture space
    light: LightSource
    light_dir_eye: Vec3f  # in eye/view space.
    texture_shape: Integer[Array, "objects 2"]
    texture_index: Integer[Array, "vertices"]
    texture_offset: Integer[Array, ""]
    texture: Texture
    specular_map: SpecularMap
    shadow: Shadow
    camera: Camera  # so accessible in FS as well.
    ambient: Colour
    diffuse: Colour
    specular: Colour


class PhongReflectionShadowTextureExtraFragmentData(NamedTuple):
    """From vertex shader to fragment shader, and from fragment shader to mixer.

    Attributes:
      - normal: in clip space, of each fragment; From VS to FS.
      - uv: in texture space, of each fragment; From VS to FS.
      - texture_index: index of texture map for each fragment; From VS to FS.
      - shadow_coord: in shadow's clip space, of each fragment; From VS to FS.
      - colour: colour when passing from FS to mixer.
    """
    normal: Vec3f = jnp.zeros(3)
    uv: Vec2f = jnp.zeros(2)
    texture_index: Integer[Array, ""] = jnp.array(0)
    shadow_coord: Vec4f = jnp.zeros(4)
    colour: Colour = jnp.zeros(3)


class PhongReflectionShadowTextureExtraMixerOutput(NamedTuple):
    """When render to only one target, for simplicity."""
    canvas: Colour


class PhongReflectionShadowTextureShader(
        Shader[PhongReflectionShadowTextureExtraInput,
               PhongReflectionShadowTextureExtraFragmentData,
               PhongReflectionShadowTextureExtraMixerOutput]):
    """Phong Shading with simple parallel lighting, texture, Phong Reflection
        approximation and ShadowMap.
    """

    @staticmethod
    @jaxtyped
    @jax.jit
    def vertex(
        gl_VertexID: ID,
        gl_InstanceID: ID,
        camera: Camera,
        extra: PhongReflectionShadowTextureExtraInput,
    ) -> tuple[PerVertex, PhongReflectionShadowTextureExtraFragmentData]:
        # Use gl_VertexID to index in `extra` buffer.
        position: Vec4f = to_homogeneous(extra.position[gl_VertexID])
        gl_Position: Vec4f = camera.to_clip(position)
        assert isinstance(gl_Position, Vec4f)

        # assume normal here is in world space. If it is in model space, it
        # must be transformed by the inverse transpose of the model matrix.
        # Ref: https://github.com/ssloy/tinyrenderer/wiki/Lesson-5:-Moving-the-camera#transformation-of-normal-vectors
        normal: Vec3f = Camera.apply_vec(
            normalise(extra.normal[gl_VertexID]),
            camera.world_to_eye_norm,
        )
        assert isinstance(normal, Vec3f)

        # shadow. Normalise here as it is not done implicitly in the pipeline.
        # the result is in shadow's clip space, as NDC.
        shadow_coord: Vec4f = normalise_homogeneous(
            extra.shadow.camera.to_clip(position))
        assert isinstance(shadow_coord, Vec4f)

        return (
            PerVertex(gl_Position=gl_Position),
            PhongReflectionShadowTextureExtraFragmentData(
                normal=normal,
                uv=extra.uv[gl_VertexID],
                texture_index=extra.texture_index[gl_VertexID],
                shadow_coord=shadow_coord,
            ),
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def interpolate(
        values: PhongReflectionShadowTextureExtraFragmentData,
        barycentric_screen: Vec3f,
        barycentric_clip: Vec3f,
    ) -> PhongReflectionShadowTextureExtraFragmentData:
        varying = Shader.interpolate(
            values=values,
            barycentric_screen=barycentric_screen,
            barycentric_clip=barycentric_clip,
        )
        # all three values should be the same, just pick the first.
        varying = varying._replace(texture_index=values.texture_index[0])
        assert isinstance(varying,
                          PhongReflectionShadowTextureExtraFragmentData)

        return varying

    @staticmethod
    @jaxtyped
    @jax.jit
    def fragment(
        gl_FragCoord: Vec4f,
        gl_FrontFacing: Bool[Array, ""],
        gl_PointCoord: Vec2f,
        varying: PhongReflectionShadowTextureExtraFragmentData,
        extra: PhongReflectionShadowTextureExtraInput,
    ) -> tuple[PerFragment, PhongReflectionShadowTextureExtraFragmentData]:
        built_in: PerFragment = Shader.fragment(
            gl_FragCoord,
            gl_FrontFacing,
            gl_PointCoord,
            varying,
            extra,
        )[0]
        assert isinstance(built_in, PerFragment)

        # shadow
        # from NDC to screen coordinates, in shadow's screen space.
        shadow_coord: Vec4f = normalise_homogeneous(
            extra.shadow.camera.viewport @ varying.shadow_coord)
        assert isinstance(shadow_coord, Vec4f)
        shadow_str: Colour = extra.shadow.strength
        assert isinstance(shadow_str, Colour)
        shadow: Colour = jnp.where(
            # if before/at shadow
            shadow_coord[2] <= extra.shadow.get(shadow_coord[:2]),
            # when not in shadow, keeps all light.
            jnp.ones_like(shadow_str),
            # if in shadow, only keep "1 - shadow_str" amount of light.
            1. - shadow_str,
        )

        # texture
        texture_index = varying.texture_index.astype(int)
        texture_shape = extra.texture_shape[texture_index]
        uv = MergedModel.uv_repeat(
            uv=varying.uv,
            shape=texture_shape,
            map_index=texture_index,
            offset=extra.texture_offset,
        )
        uv = lax.floor(uv).astype(int)
        assert isinstance(uv, Vec2i)
        texture_colour: Colour = extra.texture[uv[0], uv[1]]

        normal: Vec3f = normalise(varying.normal)
        # in Phong shading, the light direction is towards the light source;
        light_dir: Vec3f = normalise(extra.light_dir_eye)

        # Phong Reflection Model
        diffuse: float = jnp.maximum(lax.dot(normal, light_dir), 0)
        # If using standard reflection formula
        # `light_dir - 2 * diffuse * normal`, need to use
        # `-1 * light_dir` as `light_dir` instead.
        reflected_light: Vec3f = normalise(2 * lax.dot(normal, light_dir) *
                                           normal - light_dir)
        assert isinstance(reflected_light, Vec3f)

        specular: float = lax.pow(
            lax.max(reflected_light[2], 0.),
            extra.specular_map[uv[0], uv[1]],
        )

        # compute colour
        colour: Colour = (
            extra.ambient * texture_colour + shadow *
            (extra.diffuse * diffuse + extra.specular * specular) *
            texture_colour * extra.light.colour)

        return (
            PerFragment(
                keeps=jnp.logical_and(built_in.keeps, gl_FrontFacing),
                use_default_depth=built_in.use_default_depth,
            ),
            PhongReflectionShadowTextureExtraFragmentData(
                colour=lax.cond(
                    (colour >= 0).all(),
                    lambda: colour,
                    lambda: jnp.zeros(3),
                ),
                uv=varying.uv,
                normal=varying.normal,
            ),
        )

    @staticmethod
    @jaxtyped
    @jax.jit
    def mix(
        gl_FragDepth: Float[Array, "primitives"],
        keeps: Bool[Array, "primitives"],
        extra: PhongReflectionShadowTextureExtraFragmentData,
    ) -> tuple[MixerOutput, PhongReflectionShadowTextureExtraMixerOutput]:
        mixer_output: MixerOutput
        extra_output: PhongReflectionShadowTextureExtraFragmentData
        mixer_output, extra_output = Shader.mix(gl_FragDepth, keeps, extra)
        assert isinstance(mixer_output, MixerOutput)
        assert isinstance(extra_output,
                          PhongReflectionShadowTextureExtraFragmentData)

        return (
            mixer_output,
            PhongReflectionShadowTextureExtraMixerOutput(
                canvas=extra_output.colour),
        )



from functools import partial
from typing import NamedTuple

import jax
import jax.lax as lax
import jax.numpy as jnp
from jaxtyping import Array, Float, jaxtyped

from .geometry import Camera, View, Viewport
from .pipeline import render
from .shaders.depth import DepthExtraInput, DepthShader
from .types import (Buffers, Colour, FaceIndices, Vec2f, Vec2i, Vec3f,
                    Vertices, ZBuffer)


class Shadow(NamedTuple):
    """Shadow map for one light source."""
    shadow_map: ZBuffer
    """Depth map of the scene from the light source's point of view."""
    strength: Colour
    """Strength of shadow. Must be in [0, 1]. 0 means no shadow, 1 means fully
    black shadow. (1 - strength) of the original colour will be added to the
    shadowed colour.
    """
    camera: Camera
    """Camera from world space to shadow map's screen space."""

    @staticmethod
    @jaxtyped
    @partial(jax.jit, donate_argnums=(0, ))
    def render_shadow_map(
        shadow_map: ZBuffer,
        verts: Vertices,
        faces: FaceIndices,
        light_direction: Vec3f,
        viewport_matrix: Viewport,
        centre: Vec3f,
        up: Vec3f,
        strength: Colour,
        offset: float = 0.001,
        distance: float = 10.,
    ) -> "Shadow":
        """Render shadow map from light source's point of view.

        Parameters:
          - shadow_map: ZBuffer to store the depth map.
          - verts: vertices of the object.
          - faces: face indices of the object.
          - light_direction: direction of **parallel** light source, where it
            goes towards, in world space.
          - viewport_matrix: viewport matrix for rendering the objects.
          - centre: centre of the scene, same as object's camera's centre.
          - up: up direction of the scene, same as object's camera's up.
          - strength: strength of shadow. For details, see `Shadow.strength`.
          - offset: Offset to avoid self-shadowing / z-fighting. This will be
            added to the shadow map, making the shadows further away from
            the light.
          - distance: Distance from the light source to the centre of the
            scene. This is mainly to avoid objects being clipped.

        Returns: Updated `Shadow` object with shadow_map updated.
        """

        view: View = Camera.view_matrix(
            # keep "forward = -light_direction"
            eye=centre + light_direction * distance,
            centre=centre,
            up=up,
        )
        assert isinstance(view, View)

        _camera: Camera = Camera.create(
            view=view,
            projection=Camera.orthographic_projection_matrix(
                left=-1.,
                right=1.,
                bottom=-1.,
                top=1.,
                z_near=-1.,
                z_far=1.,
            ),
            viewport=viewport_matrix,
        )
        assert isinstance(_camera, Camera)

        buffers = Buffers(zbuffer=shadow_map, targets=tuple())
        extra = DepthExtraInput(position=verts)
        shadow_map, _ = render(_camera, DepthShader, buffers, faces, extra)
        shadow_map = shadow_map + offset
        assert isinstance(shadow_map, ZBuffer)

        shadow: Shadow = Shadow(
            shadow_map=shadow_map,
            strength=strength,
            camera=_camera,
        )

        return shadow

    @jaxtyped
    @partial(jax.jit, inline=True)
    def get(self, position: Vec2f) -> Float[Array, ""]:
        """Get shadow depth at `position`.

        Parameters:
          - position: position in shadow buffer's screen space.
        """
        assert isinstance(position, Vec2f), f"{position} is not a Vec3f."

        pos: Vec2i = lax.round(position[:2]).astype(int)
        assert isinstance(pos, Vec2i)

        value: Float[Array, ""] = lax.cond(
            jnp.logical_or(
                pos < 0,
                pos >= jnp.asarray(self.shadow_map.shape[:2]),
            ).any(),
            lambda: jnp.inf,  # outside shadow map, no shadow
            lambda: self.shadow_map[pos[0], pos[1]],
        )
        assert isinstance(value, Float[Array, ""])

        return value



from functools import partial
from typing import Union

import jax
import jax.numpy as jnp
from jaxtyping import Array, Bool, Integer, jaxtyped


@jaxtyped
@partial(jax.jit, inline=True)
def index_in_bound(
    indices: Integer[Array, "*any"],
    bound: Union[Integer[Array, ""], Integer[Array, "2"]],
) -> Bool[Array, ""]:
    """Check if indices are in bound.

    Parameters:
      - indices: indices to check, in any shape.
      - bound: bound to check against, assumed to be [min, max)
        (half-open interval) or [0, max) if only one value is given.
    """
    bound = jnp.asarray(bound).flatten()
    _min: int
    _max: int
    if bound.size == 2:
        _min, _max = bound
    else:
        _min, _max = 0, bound[0]

    return jnp.logical_and(
        (indices >= _min).all(),
        (indices < _max).all(),
    )
